__all__ = [
    "AbstractNode",
    "AbstractChildren",
    "AstNode",
    "AstChildren",
    "AstRoot",
    "AstCommand",
    "AstCommandSentinel",
    "AstString",
    "AstBool",
    "AstNumber",
    "AstUUID",
    "AstCoordinate",
    "AstLiteral",
    "AstOption",
    "AstWord",
    "AstGreedy",
    "AstObjective",
    "AstObjectiveCriteria",
    "AstScoreboardOperation",
    "AstTeam",
    "AstPlayerName",
    "AstLegacyScoreboardSlot",
    "AstScoreboardSlot",
    "AstSwizzle",
    "AstAdvancementPredicate",
    "AstWildcard",
    "AstColor",
    "AstColorReset",
    "AstHexColor",
    "AstSortOrder",
    "AstGamemode",
    "AstHeightmap",
    "AstEntityAnchor",
    "AstTemplateRotation",
    "AstTemplateMirror",
    "AstVector2",
    "AstVector3",
    "AstJson",
    "AstJsonValue",
    "AstJsonArray",
    "AstJsonObjectKey",
    "AstJsonObjectEntry",
    "AstJsonObject",
    "AstNbt",
    "AstNbtValue",
    "AstNbtBool",
    "AstNbtList",
    "AstNbtCompoundKey",
    "AstNbtCompoundEntry",
    "AstNbtCompound",
    "AstNbtByteArray",
    "AstNbtIntArray",
    "AstNbtLongArray",
    "AstResourceLocation",
    "AstBlockState",
    "AstBlock",
    "AstItemComponent",
    "AstItemRemovedDefaultComponent",
    "AstItemStack",
    "AstItemPredicateTestComponent",
    "AstItemPredicateTestPredicate",
    "AstItemPredicateAlternatives",
    "AstItemPredicate",
    "AstItemSlot",
    "AstItemSlots",
    "AstRange",
    "AstTime",
    "AstSelectorScoreMatch",
    "AstSelectorScores",
    "AstSelectorAdvancementPredicateMatch",
    "AstSelectorAdvancementMatch",
    "AstSelectorArgument",
    "AstSelector",
    "AstMessageText",
    "AstMessage",
    "AstNbtPathKey",
    "AstNbtPathSubscript",
    "AstNbtPath",
    "AstParticleParameters",
    "AstDustParticleParameters",
    "AstDustColorTransitionParticleParameters",
    "AstBlockParticleParameters",
    "AstFallingDustParticleParameters",
    "AstItemParticleParameters",
    "AstVibrationParticleParameters",
    "AstBlockMarkerParticleParameters",
    "AstParticle",
    "AstMacroLine",
    "AstMacroLineText",
    "AstMacroLineVariable",
    "COLORS",
]


import re
from dataclasses import dataclass, fields
from itertools import islice, permutations, zip_longest
from typing import (
    Any,
    ClassVar,
    Iterable,
    Iterator,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from uuid import UUID

from beet.core.utils import JsonDict, extra_field, required_field
from nbtlib import Byte, ByteArray, Compound, CompoundMatch, Double, Int, IntArray
from nbtlib import List as ListTag
from nbtlib import ListIndex, LongArray, NamedKey, Numeric, Path, String
from tokenstream import UNKNOWN_LOCATION, SourceLocation, set_location

from .utils import string_to_number

T = TypeVar("T")
AbstractNodeType = TypeVar("AbstractNodeType", bound="AbstractNode", covariant=True)
AstNodeType = TypeVar("AstNodeType", bound="AstNode", covariant=True)
AstLiteralType = TypeVar("AstLiteralType", bound="AstLiteral")


COLORS: Tuple[str, ...] = (
    "black",
    "dark_blue",
    "dark_green",
    "dark_aqua",
    "dark_red",
    "dark_purple",
    "gold",
    "gray",
    "dark_gray",
    "blue",
    "green",
    "aqua",
    "red",
    "light_purple",
    "yellow",
    "white",
)


@dataclass(frozen=True, slots=True)
class AbstractNode:
    """Abstract node."""

    location: SourceLocation = extra_field(default=UNKNOWN_LOCATION)
    end_location: SourceLocation = extra_field(default=UNKNOWN_LOCATION)

    def __iter__(self) -> Iterator["AbstractNode"]:
        for f in fields(self):
            attribute = getattr(self, f.name)

            if isinstance(attribute, AbstractChildren):
                yield from attribute

            if isinstance(attribute, AbstractNode):
                yield attribute

    def walk(self) -> Iterator["AbstractNode"]:
        yield self

        for f in fields(self):
            attribute = getattr(self, f.name)

            if isinstance(attribute, AbstractChildren):
                for child in attribute:
                    if isinstance(child, AbstractNode):
                        yield from child.walk()

            elif isinstance(attribute, AbstractNode):
                yield from attribute.walk()

    def dump(
        self,
        prefix: str = "",
        shallow: bool = False,
        exclude: Optional[Set[str]] = None,
    ) -> str:
        """Return a pretty-printed representation of the ast."""
        return f"{prefix}{self.__class__}\n" + "\n".join(
            f"{prefix}  {f.name}:"
            + (
                "\n"
                + (
                    "\n".join(
                        (
                            f"{prefix}    {type(child)}"
                            if shallow
                            else child.dump(prefix + "    ", shallow, exclude)
                        )
                        for child in attribute
                    )
                    if attribute
                    else prefix + "    <empty>"
                )
                if isinstance(attribute := getattr(self, f.name), AbstractChildren)
                else (
                    "\n"
                    + (
                        f"{prefix}    {type(attribute)}"
                        if shallow
                        else attribute.dump(prefix + "    ", shallow, exclude)
                    )
                    if isinstance(attribute, AbstractNode)
                    else f" {attribute!r}"
                )
            )
            for f in fields(self)
            if not exclude or f.name not in exclude
        )

    def emit_error(self, exc: T) -> T:
        """Add location information to invalid syntax exceptions."""
        return set_location(exc, self)


class AbstractChildren(Tuple[AbstractNodeType, ...]):
    """Abstract children."""

    def __new__(
        cls,
        children: Iterable[
            None | AbstractNodeType | "AbstractChildren[AbstractNodeType]"
        ] = (),
    ):
        children = [
            c
            for child in children
            if child is not None
            for c in (child if isinstance(child, AbstractChildren) else [child])
        ]
        return super().__new__(cls, children)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({super().__repr__()})"


class AstNode(AbstractNode):
    """Base class for all ast nodes."""

    parser: ClassVar[Optional[str]] = None
    compile_hints: ClassVar[JsonDict] = {}

    def __iter__(self) -> Iterator["AstNode"]:
        return super().__iter__()  # pyright: ignore[reportReturnType]

    def walk(self) -> Iterator["AstNode"]:
        return super().walk()  # pyright: ignore[reportReturnType]


class AstChildren(AbstractChildren[AstNodeType]):
    """Specialized tuple subclass for holding multiple child ast nodes."""


@dataclass(frozen=True, slots=True)
class AstRoot(AstNode):
    """Ast root node."""

    commands: AstChildren["AstCommand"] = required_field()

    parser = "root"


@dataclass(frozen=True, slots=True)
class AstCommand(AstNode):
    """Ast command node."""

    identifier: str = required_field()
    arguments: AstChildren[AstNode] = required_field()

    parser = "command"


@dataclass(frozen=True, slots=True)
class AstCommandSentinel(AstCommand):
    """Ast command sentinel node."""

    identifier: str = "mecha:sentinel"
    arguments: AstChildren[AstNode] = AstChildren()

    parser = None


@dataclass(frozen=True, slots=True)
class AstString(AstNode):
    """Ast string node."""

    value: str = required_field()

    parser = "phrase"

    @classmethod
    def from_value(cls, value: Any) -> "AstString":
        """Return a string node from the given value."""
        return AstString(value=str(value))


@dataclass(frozen=True, slots=True)
class AstBool(AstNode):
    """Ast bool node."""

    value: bool = required_field()

    parser = "bool"

    @classmethod
    def from_value(cls, value: Any) -> "AstBool":
        """Return a bool node from the given value."""
        return AstBool(value=bool(value))


@dataclass(frozen=True, slots=True)
class AstNumber(AstNode):
    """Ast number node."""

    value: Union[int, float] = required_field()

    parser = "numeric"

    @classmethod
    def from_value(cls, value: Any) -> "AstNumber":
        """Return a number node from the given value."""
        if isinstance(value, str):
            value = string_to_number(value)
        if isinstance(value, (int, float)):
            return AstNumber(value=value)
        raise ValueError(f"Invalid number {value!r}.")


@dataclass(frozen=True, slots=True)
class AstUUID(AstNode):
    """Ast uuid node."""

    value: UUID = required_field()

    parser = "uuid"

    @classmethod
    def from_value(cls, value: Any) -> "AstUUID":
        """Return a uuid node from the given value."""
        if isinstance(value, str):
            a, b, c, d, e = value.split("-")
            value = UUID(f"{a:>08}-{b:>04}-{c:>04}-{d:>04}-{e:>012}")
        if isinstance(value, UUID):
            return AstUUID(value=value)
        raise ValueError(f"Invalid UUID {value!r}.")


@dataclass(frozen=True, slots=True)
class AstCoordinate(AstNode):
    """Coordinate ast node."""

    type: Literal["absolute", "relative", "local"] = "absolute"
    value: Union[int, float] = required_field()

    parser = "coordinate"

    @classmethod
    def from_value(cls, value: Any) -> "AstCoordinate":
        """Return a coordinate node from the given value."""
        type = "absolute"
        if isinstance(value, str):
            if value.startswith("~"):
                type = "relative"
                value = value[1:]
            elif value.startswith("^"):
                type = "local"
                value = value[1:]
            value = string_to_number(value) if value else 0
        if isinstance(value, (int, float)):
            return AstCoordinate(type=type, value=value)
        raise ValueError(f"Invalid coordinate {value!r}.")


@dataclass(frozen=True, slots=True)
class AstVector2(AstNode):
    """Vector 2 ast node."""

    x: AstCoordinate = required_field()
    y: AstCoordinate = required_field()

    @classmethod
    def from_value(cls, value: Any) -> "AstVector2":
        """Return a vector 2 node from the given value."""
        if isinstance(value, str):
            value = value.split()
        x, y = map(AstCoordinate.from_value, value)
        return AstVector2(x=x, y=y)


@dataclass(frozen=True, slots=True)
class AstVector3(AstNode):
    """Vector 3 ast node."""

    x: AstCoordinate = required_field()
    y: AstCoordinate = required_field()
    z: AstCoordinate = required_field()

    @classmethod
    def from_value(cls, value: Any) -> "AstVector3":
        """Return a vector 3 node from the given value."""
        if isinstance(value, str):
            value = value.split()
        x, y, z = map(AstCoordinate.from_value, value)
        return AstVector3(x=x, y=y, z=z)


@dataclass(frozen=True, slots=True)
class AstLiteral(AstNode):
    """Base ast node for literals."""

    value: str = required_field()

    regex: ClassVar["re.Pattern[str]"] = re.compile(r"[^#:\s]+")

    @classmethod
    def from_value(cls: Type[AstLiteralType], value: Any) -> AstLiteralType:
        """Return a literal node from a given value."""
        value = str(value)
        match = cls.regex.match(value)
        if match and match[0] == value:
            return cls(value=value)
        raise ValueError(f"Invalid {cls.parser or 'literal'} {value!r}.")


@dataclass(frozen=True, slots=True)
class AstWord(AstLiteral):
    """Ast word node."""

    parser = "word"
    regex = re.compile(r"[0-9A-Za-z_\.\+\-]+")


@dataclass(frozen=True, slots=True)
class AstOption(AstLiteral):
    """Base node for options."""

    options: ClassVar[Set[str]] = set()

    def __init_subclass__(cls):
        patterns = [
            rf"{pattern}\b" if pattern[-1].isalnum() else pattern
            for option in sorted(cls.options, key=lambda s: (-len(s), s))
            if (pattern := re.escape(option))
        ]
        cls.regex = re.compile("|".join(patterns))


@dataclass(frozen=True, slots=True)
class AstGreedy(AstLiteral):
    """Ast greedy node."""

    parser = "greedy"
    regex = re.compile(r".+")


@dataclass(frozen=True, slots=True)
class AstObjective(AstLiteral):
    """Ast objective node."""

    parser = "objective"
    regex = re.compile(r"[a-zA-Z0-9_.+-]+")


@dataclass(frozen=True, slots=True)
class AstObjectiveCriteria(AstLiteral):
    """Ast objective criteria node."""

    parser = "objective_criteria"
    regex = re.compile(r"[a-zA-Z0-9_.+-]+(?::[a-zA-Z0-9_.+-]+)?")


@dataclass(frozen=True, slots=True)
class AstScoreboardOperation(AstOption):
    """Ast scoreboard operation node."""

    parser = "scoreboard_operation"
    options = {"+=", "-=", "*=", "/=", "%=", "=", "><", "<", ">"}


@dataclass(frozen=True, slots=True)
class AstTeam(AstLiteral):
    """Ast team node."""

    parser = "team"
    regex = re.compile(r"[a-zA-Z0-9_.+-]+")


@dataclass(frozen=True, slots=True)
class AstPlayerName(AstLiteral):
    """Ast player name node."""

    parser = "player_name"
    regex = re.compile(r"(?![@*[({])\S+")


@dataclass(frozen=True, slots=True)
class AstLegacyScoreboardSlot(AstOption):
    """Ast legacy scoreboard slot node."""

    parser = "scoreboard_slot"
    options = {f"sidebar.team.{color}" for color in COLORS} | {
        "list",
        "sidebar",
        "belowName",
    }


@dataclass(frozen=True, slots=True)
class AstScoreboardSlot(AstOption):
    """Ast scoreboard slot node."""

    parser = "scoreboard_slot"
    options = {f"sidebar.team.{color}" for color in COLORS} | {
        "list",
        "sidebar",
        "below_name",
    }


@dataclass(frozen=True, slots=True)
class AstSwizzle(AstOption):
    """Ast swizzle node."""

    parser = "swizzle"
    options = (
        set("xyz")
        | set(map("".join, permutations("xyz", 2)))
        | set(map("".join, permutations("xyz", 3)))
    )


@dataclass(frozen=True, slots=True)
class AstAdvancementPredicate(AstLiteral):
    """Ast advancement predicate node."""

    parser = "advancement_predicate"
    regex = re.compile(r"[0-9A-Za-z_\.\+\-]+")


@dataclass(frozen=True, slots=True)
class AstWildcard(AstLiteral):
    """Ast wildcard node."""

    value: str = "*"

    parser = "wildcard"
    regex = re.compile(r"\*")


@dataclass(frozen=True, slots=True)
class AstColor(AstOption):
    """Ast color node."""

    parser = "color"
    options = set(COLORS)


@dataclass(frozen=True, slots=True)
class AstColorReset(AstOption):
    """Ast color reset node."""

    parser = "color_reset"
    options = {"reset"}


@dataclass(frozen=True, slots=True)
class AstHexColor(AstLiteral):
    """Ast color node."""

    parser = "hex_color"
    regex = re.compile(r"[0-9a-fA-F]{6}|[0-9a-fA-F]{3}")


@dataclass(frozen=True, slots=True)
class AstSortOrder(AstOption):
    """Ast sort order node."""

    parser = "sort_order"
    options = {"nearest", "furthest", "random", "arbitrary"}


@dataclass(frozen=True, slots=True)
class AstGamemode(AstOption):
    """Ast gamemode node."""

    parser = "gamemode"
    options = {"adventure", "creative", "spectator", "survival"}


@dataclass(frozen=True, slots=True)
class AstHeightmap(AstOption):
    """Ast heightmap node."""

    parser = "heightmap"
    options = {
        "world_surface",
        "motion_blocking",
        "motion_blocking_no_leaves",
        "ocean_floor",
    }


@dataclass(frozen=True, slots=True)
class AstEntityAnchor(AstOption):
    """Ast entity anchor node."""

    parser = "entity_anchor"
    options = {"eyes", "feet"}


@dataclass(frozen=True, slots=True)
class AstTemplateRotation(AstOption):
    """Ast template rotation node."""

    parser = "template_rotation"
    options = {"180", "clockwise_90", "counterclockwise_90", "none"}


@dataclass(frozen=True, slots=True)
class AstTemplateMirror(AstOption):
    """Ast template mirror node."""

    parser = "template_mirror"
    options = {"front_back", "left_right", "none"}


@dataclass(frozen=True, slots=True)
class AstJson(AstNode):
    """Base ast node for json."""

    parser = "json"

    def evaluate(self) -> Any:
        """Return the json value."""
        raise NotImplementedError()

    @classmethod
    def from_value(cls, value: Any) -> "AstJson":
        """Create json ast nodes representing the specified json value."""
        if value is None or isinstance(value, (bool, str, int, float)):
            return AstJsonValue(value=value)
        elif isinstance(value, Mapping):
            object: Mapping[Any, Any] = value
            return AstJsonObject(
                entries=AstChildren(
                    AstJsonObjectEntry(
                        key=AstJsonObjectKey(value=str(k)),
                        value=cls.from_value(v),
                    )
                    for k, v in object.items()
                )
            )
        elif isinstance(value, Sequence):
            array: Sequence[Any] = value
            return AstJsonArray(elements=AstChildren(cls.from_value(e) for e in array))
        else:
            raise ValueError(f"Invalid json value of type {type(value)!r}.")


@dataclass(frozen=True, slots=True)
class AstJsonValue(AstJson):
    """Ast json value node."""

    value: Any = required_field()

    parser = None

    def evaluate(self) -> Any:
        return self.value


@dataclass(frozen=True, slots=True)
class AstJsonArray(AstJson):
    """Ast json array node."""

    elements: AstChildren[AstJson] = required_field()

    parser = None

    def evaluate(self) -> Any:
        return [element.evaluate() for element in self.elements]


@dataclass(frozen=True, slots=True)
class AstJsonObjectKey(AstNode):
    """Ast json object key node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class AstJsonObjectEntry(AstNode):
    """Ast json object entry node."""

    key: AstJsonObjectKey = required_field()
    value: AstJson = required_field()


@dataclass(frozen=True, slots=True)
class AstJsonObject(AstJson):
    """Ast json object node."""

    entries: AstChildren[AstJsonObjectEntry] = required_field()

    parser = "json"

    def evaluate(self) -> Any:
        return {entry.key.value: entry.value.evaluate() for entry in self.entries}


@dataclass(frozen=True, slots=True)
class AstNbt(AstNode):
    """Base ast node for nbt."""

    parser = "nbt"

    def evaluate(self) -> Any:
        """Return the nbt value."""
        raise NotImplementedError()

    @overload
    @classmethod
    def from_value(cls, value: Union[bool, int, float, str]) -> "AstNbtValue": ...

    @overload
    @classmethod
    def from_value(cls, value: Mapping[Any, Any]) -> "AstNbtCompound": ...

    @overload
    @classmethod
    def from_value(cls, value: Sequence[Any]) -> "AstNbtList": ...

    @overload
    @classmethod
    def from_value(cls, value: Any) -> "AstNbt": ...

    @classmethod
    def from_value(cls, value: Any) -> "AstNbt":
        """Create nbt ast nodes representing the specified value."""
        if isinstance(value, (Numeric, String)):
            return AstNbtValue(value=value)
        elif isinstance(value, bool):
            return AstNbtBool(value=Byte(1 if value else 0))
        elif isinstance(value, int):
            return AstNbtValue(value=Int(value))
        elif isinstance(value, float):
            return AstNbtValue(value=Double(value))
        elif isinstance(value, str):
            return AstNbtValue(value=String(value))
        elif isinstance(value, ByteArray):
            return AstNbtByteArray(
                elements=AstChildren(cls.from_value(e) for e in value)
            )
        elif isinstance(value, IntArray):
            return AstNbtIntArray(
                elements=AstChildren(cls.from_value(e) for e in value)
            )
        elif isinstance(value, LongArray):
            return AstNbtLongArray(
                elements=AstChildren(cls.from_value(e) for e in value)
            )
        elif isinstance(value, Mapping):
            compound: Mapping[Any, Any] = value
            return AstNbtCompound(
                entries=AstChildren(
                    AstNbtCompoundEntry(
                        key=AstNbtCompoundKey(value=str(k)),
                        value=cls.from_value(v),
                    )
                    for k, v in compound.items()
                )
            )
        elif isinstance(value, Sequence):
            lst: Sequence[Any] = value
            return AstNbtList(elements=AstChildren(cls.from_value(e) for e in lst))
        else:
            raise ValueError(f"Invalid nbt value of type {type(value)!r}.")


@dataclass(frozen=True, slots=True)
class AstNbtValue(AstNbt):
    """Ast nbt value node."""

    value: Any = required_field()

    parser = None

    def evaluate(self) -> Any:
        return self.value


@dataclass(frozen=True, slots=True)
class AstNbtBool(AstNbtValue):
    """Ast nbt bool node."""


@dataclass(frozen=True, slots=True)
class AstNbtList(AstNbt):
    """Ast nbt list node."""

    elements: AstChildren[AstNbt] = required_field()

    parser = None

    def evaluate(self) -> Any:
        return ListTag([element.evaluate() for element in self.elements])


@dataclass(frozen=True, slots=True)
class AstNbtCompoundKey(AstNode):
    """Ast nbt compound key node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class AstNbtCompoundEntry(AstNode):
    """Ast nbt compound entry node."""

    key: AstNbtCompoundKey = required_field()
    value: AstNbt = required_field()


@dataclass(frozen=True, slots=True)
class AstNbtCompound(AstNbt):
    """Ast nbt compound node."""

    entries: AstChildren[AstNbtCompoundEntry] = required_field()

    parser = "nbt_compound"

    def evaluate(self) -> Any:
        return Compound(
            {entry.key.value: entry.value.evaluate() for entry in self.entries},
        )


@dataclass(frozen=True, slots=True)
class AstNbtByteArray(AstNbt):
    """Ast nbt byte array node."""

    elements: AstChildren[AstNbt] = required_field()

    parser = None
    array_prefix = "B"

    def evaluate(self) -> Any:
        return ByteArray([element.evaluate() for element in self.elements])


@dataclass(frozen=True, slots=True)
class AstNbtIntArray(AstNbt):
    """Ast nbt int array node."""

    elements: AstChildren[AstNbt] = required_field()

    parser = None
    array_prefix = "I"

    def evaluate(self) -> Any:
        return IntArray([element.evaluate() for element in self.elements])


@dataclass(frozen=True, slots=True)
class AstNbtLongArray(AstNbt):
    """Ast nbt long array node."""

    elements: AstChildren[AstNbt] = required_field()

    parser = None
    array_prefix = "L"

    def evaluate(self) -> Any:
        return LongArray([element.evaluate() for element in self.elements])


@dataclass(frozen=True, slots=True)
class AstResourceLocation(AstNode):
    """Ast resource location node."""

    is_tag: bool = False
    namespace: Optional[str] = None
    path: str = required_field()

    parser = "resource_location_or_tag"

    @classmethod
    def from_value(cls, value: Any) -> "AstResourceLocation":
        """Create a resource location node representing the given value."""
        value = str(value)
        if is_tag := value.startswith("#"):
            value = value[1:]
        namespace, _, path = value.rpartition(":")
        return cls(is_tag=is_tag, namespace=namespace or None, path=path)

    def get_value(self) -> str:
        """Return the value of the resource location as a string."""
        prefix = "#" if self.is_tag else ""
        namespace = f"{self.namespace}:" if self.namespace else ""
        return prefix + namespace + self.path

    def get_canonical_value(self) -> str:
        """Return the canonical value of the resource location as a string."""
        prefix = "#" if self.is_tag else ""
        namespace = f"{self.namespace}:" if self.namespace else "minecraft:"
        return prefix + namespace + self.path


@dataclass(frozen=True, slots=True)
class AstBlockState(AstNode):
    """Ast block state node."""

    key: AstString = required_field()
    value: AstString = required_field()


@dataclass(frozen=True, slots=True)
class AstBlock(AstNode):
    """Ast block node."""

    identifier: AstResourceLocation = required_field()
    block_states: AstChildren[AstBlockState] = AstChildren()
    data_tags: Optional[AstNbtCompound] = None


@dataclass(frozen=True, slots=True)
class AstItemComponent(AstNode):
    """Ast item component node."""

    key: AstResourceLocation = required_field()
    value: AstNbt = required_field()


@dataclass(frozen=True, slots=True)
class AstItemRemovedDefaultComponent(AstNode):
    """Ast item removed default component node."""

    key: AstResourceLocation = required_field()


@dataclass(frozen=True, slots=True)
class AstItemStack(AstNode):
    """Ast item stack node."""

    identifier: AstResourceLocation = required_field()
    arguments: AstChildren[Union[AstItemComponent, AstItemRemovedDefaultComponent]] = (
        AstChildren()
    )
    data_tags: Optional[AstNbtCompound] = None  # legacy compat


@dataclass(frozen=True, slots=True)
class AstItemPredicateTestComponent(AstNode):
    """Ast item predicate test component node."""

    inverted: bool = False
    key: AstResourceLocation = required_field()
    value: Optional[AstNbt] = None


@dataclass(frozen=True, slots=True)
class AstItemPredicateTestPredicate(AstNode):
    """Ast item predicate test predicate node."""

    inverted: bool = False
    key: AstResourceLocation = required_field()
    value: AstNbt = required_field()


@dataclass(frozen=True, slots=True)
class AstItemPredicateAlternatives(AstNode):
    """Ast item predicate alternatives node."""

    alternatives: AstChildren[
        Union[AstItemPredicateTestComponent, AstItemPredicateTestPredicate]
    ] = required_field()


@dataclass(frozen=True, slots=True)
class AstItemPredicate(AstNode):
    """Ast item predicate node."""

    identifier: Union[AstResourceLocation, AstWildcard] = required_field()
    arguments: AstChildren[AstItemPredicateAlternatives] = AstChildren()
    data_tags: Optional[AstNbtCompound] = None  # legacy compat


@dataclass(frozen=True, slots=True)
class AstItemSlot(AstOption):
    """Ast item slot node."""

    parser = "item_slot"
    options = (
        {
            "armor.body",
            "armor.chest",
            "armor.feet",
            "armor.head",
            "armor.legs",
            "contents",
            "saddle",
            "weapon",
            "weapon.mainhand",
            "weapon.offhand",
            "player.cursor",
            "player.crafting.0",
            "player.crafting.1",
            "player.crafting.2",
            "player.crafting.3",
        }
        | {f"container.{n}" for n in range(54)}
        | {f"enderchest.{n}" for n in range(27)}
        | {f"hotbar.{n}" for n in range(9)}
        | {f"inventory.{n}" for n in range(27)}
        | {"horse.saddle", "horse.chest", "horse.armor"}
        | {f"horse.{n}" for n in range(15)}
        | {f"villager.{n}" for n in range(8)}
    )


@dataclass(frozen=True, slots=True)
class AstItemSlots(AstOption):
    """Ast item slots node."""

    parser = "item_slots"
    options = AstItemSlot.options | {
        "container.*",
        "hotbar.*",
        "inventory.*",
        "enderchest.*",
        "villager.*",
        "horse.*",
        "weapon.*",
        "armor.*",
        "player.crafting.*",
    }


@dataclass(frozen=True, slots=True)
class AstRange(AstNode):
    """Ast range node."""

    min: Optional[Union[int, float]] = required_field()
    max: Optional[Union[int, float]] = required_field()

    parser = "range"

    @property
    def exact(self) -> bool:
        """Wether the range is an exact match."""
        return self.min == self.max

    @property
    def value(self) -> Union[int, float]:
        """Return the exact value."""
        assert self.min is not None
        return self.min

    @classmethod
    def from_value(
        cls,
        value: Union[
            str,
            int,
            float,
            Tuple[Union[int, float, str], Union[int, float, str]],
        ],
    ) -> "AstRange":
        """Create a range node from a given value."""
        if isinstance(value, (int, float)):
            return AstRange(min=value, max=value)

        if isinstance(value, tuple):
            min, max = value
            return AstRange(
                min=string_to_number(min) if isinstance(min, str) else min,
                max=string_to_number(max) if isinstance(max, str) else max,
            )

        minimum, separator, maximum = value.partition("..")

        if not separator:
            maximum = minimum

        return AstRange(
            min=string_to_number(minimum) if minimum else None,
            max=string_to_number(maximum) if maximum else None,
        )


@dataclass(frozen=True, slots=True)
class AstTime(AstNode):
    """Ast time node."""

    value: Union[int, float] = required_field()
    unit: Literal["day", "second", "tick"] = "tick"

    parser = "time"

    @classmethod
    def from_value(cls, value: Union[str, int, float]) -> "AstTime":
        """Create a time node from a given value."""
        unit = "tick"

        if isinstance(value, str):
            if value.endswith(("d", "s", "t")):
                if value[-1] == "d":
                    unit = "day"
                elif value[-1] == "s":
                    unit = "second"
                else:
                    unit = "tick"
                value = value[:-1]
            value = string_to_number(value)

        return cls(value=value, unit=unit)


@dataclass(frozen=True, slots=True)
class AstSelectorScoreMatch(AstNode):
    """Ast selector score match node."""

    key: AstObjective = required_field()
    value: AstRange = required_field()


@dataclass(frozen=True, slots=True)
class AstSelectorScores(AstNode):
    """Ast selector scores node."""

    scores: AstChildren[AstSelectorScoreMatch] = required_field()


@dataclass(frozen=True, slots=True)
class AstSelectorAdvancementPredicateMatch(AstNode):
    """Ast selector advancement predicate match node."""

    key: AstAdvancementPredicate = required_field()
    value: AstBool = required_field()


@dataclass(frozen=True, slots=True)
class AstSelectorAdvancementMatch(AstNode):
    """Ast selector advancement match node."""

    key: AstResourceLocation = required_field()
    value: Union[AstBool, AstChildren[AstSelectorAdvancementPredicateMatch]] = (
        required_field()
    )


@dataclass(frozen=True, slots=True)
class AstSelectorAdvancements(AstNode):
    """Ast selector advancements node."""

    advancements: AstChildren[AstSelectorAdvancementMatch] = required_field()


@dataclass(frozen=True, slots=True)
class AstSelectorArgument(AstNode):
    """Ast selector argument node."""

    inverted: bool = False
    key: AstString = required_field()
    value: Optional[AstNode] = required_field()


@dataclass(frozen=True, slots=True)
class AstSelector(AstNode):
    """Ast selector node."""

    variable: Literal["p", "r", "a", "e", "s", "n"] = required_field()
    arguments: AstChildren[AstSelectorArgument] = AstChildren()

    parser = "selector"


@dataclass(frozen=True, slots=True)
class AstMessageText(AstLiteral):
    """Ast message text node."""

    value: str = required_field()
    regex = re.compile(r".+")


@dataclass(frozen=True, slots=True)
class AstMessage(AstNode):
    """Ast message node."""

    fragments: AstChildren[Union[AstMessageText, AstSelector]] = required_field()

    @classmethod
    def from_value(cls, obj: Any) -> "AstMessage":
        """Return message node from a given value."""
        return AstMessage(fragments=AstChildren([AstMessageText.from_value(obj)]))


@dataclass(frozen=True, slots=True)
class AstNbtPathKey(AstNode):
    """Ast nbt path key node."""

    value: str = required_field()


@dataclass(frozen=True, slots=True)
class AstNbtPathSubscript(AstNode):
    """Ast nbt path subscript node."""

    index: Union[None, AstNumber, AstNbtCompound] = None


@dataclass(frozen=True, slots=True)
class AstNbtPath(AstNode):
    """Ast nbt path node."""

    components: AstChildren[
        Union[AstNbtPathKey, AstNbtCompound, AstNbtPathSubscript]
    ] = required_field()

    parser = "nbt_path"

    @classmethod
    def from_value(cls, value: Any) -> "AstNbtPath":
        """Create nbt path ast nodes representing the specified value."""
        if isinstance(value, str):
            value = Path(value)
        if not isinstance(value, Path):
            raise ValueError(f"Invalid nbt path value of type {type(value)!r}.")

        accessors = zip_longest(value, islice(value, 1, None))
        components: List[Union[AstNbtPathKey, AstNbtCompound, AstNbtPathSubscript]] = []

        for accessor, next_accessor in accessors:
            if isinstance(accessor, NamedKey):
                components.append(AstNbtPathKey(value=accessor.key))

            elif isinstance(accessor, ListIndex):
                index = (
                    AstNumber(value=accessor.index)
                    if isinstance(accessor.index, int)
                    else None
                )

                # Special-case for the way nbtlib.Path stores compound subscripts.
                if index is None and isinstance(next_accessor, CompoundMatch):
                    compound: Mapping[Any, Any] = next_accessor.compound
                    index = AstNbt.from_value(compound)
                    next(accessors, None)

                components.append(AstNbtPathSubscript(index=index))

            elif isinstance(accessor, CompoundMatch):
                compound: Mapping[Any, Any] = accessor.compound
                components.append(AstNbt.from_value(compound))

        if not components:
            raise ValueError("Empty nbt path not allowed.")
        return AstNbtPath(components=AstChildren(components))


@dataclass(frozen=True, slots=True)
class AstParticleParameters(AstNode):  # legacy compat
    """Base ast node for particle parameters."""


@dataclass(frozen=True, slots=True)
class AstDustParticleParameters(AstParticleParameters):
    """Ast dust particle parameters node."""

    red: AstNumber = required_field()
    green: AstNumber = required_field()
    blue: AstNumber = required_field()
    size: AstNumber = required_field()


@dataclass(frozen=True, slots=True)
class AstDustColorTransitionParticleParameters(AstParticleParameters):
    """Ast dust color transition particle parameters node."""

    red: AstNumber = required_field()
    green: AstNumber = required_field()
    blue: AstNumber = required_field()
    size: AstNumber = required_field()
    end_red: AstNumber = required_field()
    end_green: AstNumber = required_field()
    end_blue: AstNumber = required_field()


@dataclass(frozen=True, slots=True)
class AstBlockParticleParameters(AstParticleParameters):
    """Ast block particle parameters node."""

    block: AstBlock = required_field()


@dataclass(frozen=True, slots=True)
class AstFallingDustParticleParameters(AstParticleParameters):
    """Ast falling dust particle parameters node."""

    block: AstBlock = required_field()


@dataclass(frozen=True, slots=True)
class AstItemParticleParameters(AstParticleParameters):
    """Ast item particle parameters node."""

    item: AstItemStack = required_field()


@dataclass(frozen=True, slots=True)
class AstVibrationParticleParameters(AstParticleParameters):
    """Ast vibration particle parameters node."""

    destination_x: AstNumber = required_field()
    destination_y: AstNumber = required_field()
    destination_z: AstNumber = required_field()
    duration: AstNumber = required_field()


@dataclass(frozen=True, slots=True)
class AstBlockMarkerParticleParameters(AstParticleParameters):
    """Ast block marker particle parameters node."""

    block: AstBlock = required_field()


@dataclass(frozen=True, slots=True)
class AstSculkChargeParticleParameters(AstParticleParameters):
    """Ast sculk charge particle parameters node."""

    angle: AstNumber = required_field()


@dataclass(frozen=True, slots=True)
class AstShriekParticleParameters(AstParticleParameters):
    """Ast shriek particle parameters node."""

    delay: AstNumber = required_field()


@dataclass(frozen=True, slots=True)
class AstParticle(AstNode):
    """Ast particle node."""

    name: AstResourceLocation = required_field()
    parameters: Optional[Union[AstParticleParameters, AstNbtCompound]] = None

    parser = "particle"


@dataclass(frozen=True, slots=True)
class AstMacroLineText(AstLiteral):
    """Ast macro line text node."""

    parser = "macro_line_text"
    regex = re.compile(r"(?:[^\n$]|\$(?!\())+")


@dataclass(frozen=True, slots=True)
class AstMacroLineVariable(AstLiteral):
    """Ast macro line variable node."""

    parser = "macro_line_variable"
    regex = re.compile(r"[0-9A-Za-z_]+")


@dataclass(frozen=True, slots=True)
class AstMacroLine(AstCommandSentinel):
    """Ast macro line node."""

    parser = "macro_line"

    compile_hints = {"skip_execute_inline_single_command": True}



__all__ = [
    "CommandTree",
]


import json
from importlib.resources import files
from pathlib import Path
from typing import Dict, Iterator, List, Literal, Optional, Set, Tuple, Union

from beet import ErrorMessage
from beet.core.utils import FileSystemPath, JsonDict, VersionNumber, split_version
from pydantic import BaseModel


class CommandTree(BaseModel):
    """Deserialized representation of the command tree."""

    type: Literal["root", "literal", "argument"]
    parser: Optional[str] = None
    properties: Optional[JsonDict] = None
    executable: Optional[bool] = None
    redirect: Optional[Tuple[str, ...]] = None
    children: Optional[Dict[str, "CommandTree"]] = None
    subcommand: bool = False

    @classmethod
    def load_from(
        cls,
        *args: FileSystemPath,
        version: Optional[VersionNumber] = None,
        unpatched: bool = False,
        patch_only: bool = False,
    ) -> "CommandTree":
        """Load the command tree from a file."""
        sources: List[str] = []

        for filename in args:
            sources.append(Path(filename).read_text())

        version = split_version(version) if version is not None else None

        if version and not patch_only:
            version_name = "_".join(map(str, version))
            try:
                sources.append(
                    files("mecha.resources")
                    .joinpath(f"{version_name}.json")
                    .read_text()
                )
            except FileNotFoundError as exc:
                raise ErrorMessage(f"Invalid minecraft version {version!r}.") from exc

        tree = cls.model_validate_json(sources[0])

        for source in sources[1:]:
            tree.extend(cls.model_validate_json(source))

        if version and not unpatched:
            for patch in json.loads(
                files("mecha.resources").joinpath("patches.json").read_text()
            ):
                if any(split_version(v) == version for v in patch["versions"]):
                    tree.extend(cls(**patch["config"]))

        return tree

    def get(self, *scope: Union[str, Tuple[str, ...], None]) -> Optional["CommandTree"]:
        """Retrieve a nested node."""
        node = self
        for arg in scope:
            if arg is None:
                return None
            for name in [arg] if isinstance(arg, str) else arg:
                if node.children:
                    node = node.children.get(name)
                    if not node:
                        return None
        return node

    def get_literal(self, name: str) -> Optional["CommandTree"]:
        """Retrieve a literal child by name."""
        if self.children:
            if child := self.children.get(name):
                if child.type == "literal":
                    return child
        return None

    def get_argument(self, name: str) -> Optional["CommandTree"]:
        """Retrieve an argument child by name."""
        if self.children:
            if child := self.children.get(name):
                if child.type == "argument":
                    return child
        return None

    def get_all_literals(self) -> Iterator[Tuple[str, "CommandTree"]]:
        """Retrieve all the literal children."""
        if self.children:
            for name, child in self.children.items():
                if child.type == "literal":
                    yield name, child

    def get_all_arguments(self) -> Iterator[Tuple[str, "CommandTree"]]:
        """Retrieve all the argument children."""
        if self.children:
            for name, child in self.children.items():
                if child.type == "argument":
                    yield name, child

    def extend(self, other: "CommandTree") -> "CommandTree":
        """Merge nodes from another command tree."""
        self.type = other.type

        if other.parser is not None:
            self.parser = other.parser

        if other.properties is not None:
            self.properties = other.properties

        if other.executable is not None:
            self.executable = other.executable

        if other.redirect is not None:
            self.redirect = other.redirect

        if other.children:
            if self.children is None:
                self.children = {}
            for name, child in other.children.items():
                if name in self.children:
                    self.children[name].extend(child)
                else:
                    self.children[name] = child

        return self

    def resolve(
        self,
        root: Optional["CommandTree"] = None,
        scope: Tuple[str, ...] = (),
    ) -> "CommandTree":
        """Resolve redirections and subcommands."""
        if not root:
            root = self

        hoisted: Set[str] = set()

        if self.redirect is not None:
            if target := root.get(self.redirect):
                if scope[: len(self.redirect)] == self.redirect:
                    target.subcommand = True
                if target.children:
                    if not self.children:
                        self.children = {}
                    self.children.update(target.children)
                    hoisted.update(target.children)
            else:
                raise ValueError(f"Invalid redirect {self.redirect}.")

        if self.children:
            for name, child in self.children.items():
                if name not in hoisted:
                    child.resolve(root, scope + (name,))

        if self.redirect is None and not self.children and not self.executable:
            self.redirect = ()
            self.resolve(root, scope)

        return self

    def __repr__(self) -> str:
        args = ", ".join(
            f"{name}="
            + (
                "{" + ", ".join(f"{k!r}: ..." for k in value) + "}"
                if name == "children"
                else repr(value)
            )
            for name in self.__class__.model_fields
            if (value := getattr(self, name))
        )
        return f"{self.__class__.__name__}({args})"

    def __str__(self) -> str:
        return self.__repr__()


CommandTree.model_rebuild()



"""Plugin for inserting raw commands."""

__all__ = [
    "RawCommandSerializer",
]


from typing import List

from beet import Context

from mecha import AstCommand, AstGreedy, Mecha, Visitor, rule

RAW_COMMAND_TREE = {
    "type": "root",
    "children": {
        "raw": {
            "type": "literal",
            "children": {
                "value": {
                    "type": "argument",
                    "executable": True,
                    "parser": "brigadier:string",
                    "properties": {
                        "type": "greedy",
                    },
                }
            },
        },
    },
}


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    mc.spec.add_commands(RAW_COMMAND_TREE)
    mc.serialize.extend(RawCommandSerializer())


class RawCommandSerializer(Visitor):
    @rule(AstCommand, identifier="raw:value")
    def raw_command(self, node: AstCommand, result: List[str]):
        if isinstance(command := node.arguments[0], AstGreedy):
            result.append(command.value)



__all__ = [
    "Parser",
    "CommandSpec",
]


from dataclasses import dataclass
from typing import Any, Dict, Protocol, Set, Tuple, Union

from beet import LATEST_MINECRAFT_VERSION
from beet.core.utils import JsonDict, extra_field
from tokenstream import TokenStream

from .config import CommandTree
from .prototype import CommandArgument, CommandPrototype, CommandSignature


class Parser(Protocol):
    """Protocol describing parser signature."""

    def __call__(self, stream: TokenStream) -> Any: ...


@dataclass
class CommandSpec:
    """Class responsible for managing the command specification."""

    multiline: bool = False

    tree: CommandTree = extra_field(
        default_factory=lambda: CommandTree.load_from(version=LATEST_MINECRAFT_VERSION)
    )
    prototypes: Dict[str, CommandPrototype] = extra_field(default_factory=dict)

    parsers: Dict[str, Parser] = extra_field(default_factory=dict)

    def __post_init__(self):
        self.update()

    def add_commands(self, tree: Union[CommandTree, JsonDict]):
        """Extend the command tree and regenerate prototypes."""
        if not isinstance(tree, CommandTree):
            tree = CommandTree(**tree)
        self.tree.extend(tree)
        self.update()

    def update(self):
        """Recalculate command prototypes."""
        self.tree.resolve()

        self.prototypes.clear()
        self.generate_prototypes(self.tree)

    def generate_prototypes(
        self,
        tree: CommandTree,
        scope: Tuple[str, ...] = (),
        signature: CommandSignature = (),
        arguments: Tuple[int, ...] = (),
    ):
        """Traverse command tree and generate prototypes."""
        if tree.executable:
            identifier = ":".join(scope)
            self.prototypes[identifier] = CommandPrototype(
                identifier, signature, arguments
            )

        recursive = False
        hoisted: Set[str] = set()

        if target := self.tree.get(tree.redirect):
            recursive = target.subcommand
            if recursive and target.children:
                hoisted.update(target.children)

        if scope and tree.subcommand or recursive:
            identifier = ":".join(scope + ("subcommand",))
            argument_scope = tree.redirect if tree.redirect is not None else scope
            self.prototypes[identifier] = CommandPrototype(
                identifier,
                signature + (CommandArgument("subcommand", argument_scope),),
                arguments + (len(signature),),
            )
            signature = signature[len(scope) :]
            arguments = tuple(
                offset for arg in arguments if (offset := arg - len(scope)) >= 0
            )

        if tree.children:
            for name, child in tree.children.items():
                if name in hoisted:
                    continue
                elif child.type == "literal":
                    self.generate_prototypes(
                        child,
                        scope + (name,),
                        signature + (name,),
                        arguments,
                    )
                elif child.type == "argument":
                    self.generate_prototypes(
                        child,
                        scope + (name,),
                        signature + (CommandArgument(name, scope + (name,)),),
                        arguments + (len(signature),),
                    )



__all__ = [
    "CommandSignature",
    "CommandPrototype",
    "CommandArgument",
]


from typing import NamedTuple, Tuple, Union

CommandSignature = Tuple[Union[str, "CommandArgument"], ...]


class CommandArgument(NamedTuple):
    """Class representing an argument in a command prototype."""

    name: str
    scope: Tuple[str, ...]


class CommandPrototype(NamedTuple):
    """Class representing a command prototype."""

    identifier: str
    signature: CommandSignature
    arguments: Tuple[int, ...]

    def get_argument(self, arg: Union[str, int]) -> CommandArgument:
        """Return the argument corresponding to the given name or index."""
        if isinstance(arg, str):
            for i in range(len(self.arguments)):
                argument = self.get_argument(i)
                if argument.name == arg:
                    return argument
            else:
                raise ValueError(f"No argument {arg!r}.")

        return self.signature[self.arguments[arg]]  # pyright: ignore[reportReturnType]

    def usage(self) -> str:
        """Return a string showing the command usage."""
        return " ".join(
            arg if isinstance(arg, str) else f"<{arg.name}>" for arg in self.signature
        )



"""Plugin for declaring function tags inline."""

__all__ = [
    "InlineFunctionTagHandler",
]


import logging
from dataclasses import dataclass, replace
from importlib.resources import files
from typing import List

from beet import Context, FunctionTag, Generator
from beet.core.utils import required_field
from tokenstream import set_location

from mecha import (
    AstChildren,
    AstCommand,
    AstResourceLocation,
    AstRoot,
    CommandTree,
    CompilationDatabase,
    Diagnostic,
    Mecha,
    Visitor,
    rule,
)

logger = logging.getLogger(__name__)


def beet_default(ctx: Context):
    logger.warning('Deprecated in favor of "mecha.contrib.nested_resources".')

    mc = ctx.inject(Mecha)

    commands_json = (
        files("mecha.resources").joinpath("inline_function_tag.json").read_text()
    )
    mc.spec.add_commands(CommandTree.model_validate_json(commands_json))

    mc.transform.extend(inline_execute_function_tag)
    mc.steps.insert(
        mc.steps.index(mc.transform) + 1,
        InlineFunctionTagHandler(generate=ctx.generate, database=mc.database),
    )


@rule(AstCommand, identifier="execute:run:subcommand")
def inline_execute_function_tag(node: AstCommand):
    if isinstance(command := node.arguments[0], AstCommand):
        if command.identifier == "function:tag:name":
            d = Diagnostic("error", "Can't add function tag with execute.")
            yield set_location(d, command.arguments[0])
    return node


@dataclass
class InlineFunctionTagHandler(Visitor):
    """Handler for inline function tags."""

    generate: Generator = required_field()
    database: CompilationDatabase = required_field()

    @rule(AstRoot)
    def inline_function_tag(self, node: AstRoot):
        changed = False
        commands: List[AstCommand] = []

        for command in node.commands:
            if command.identifier == "function:tag:name" and isinstance(
                tag := command.arguments[0], AstResourceLocation
            ):
                changed = True
                if path := self.database[self.database.current].resource_location:
                    self.generate(
                        tag.get_canonical_value(), merge=FunctionTag({"values": [path]})
                    )
                else:
                    d = Diagnostic("error", "No current path to add function tag.")
                    yield set_location(d, tag)
                    return node
            else:
                commands.append(command)

        if changed:
            return replace(node, commands=AstChildren(commands))

        return node



"""Plugin for handling lexically nested resource locations."""

__all__ = [
    "AstNestedLocation",
    "UnresolvedNestedLocation",
    "nested_location",
    "parse_nested_location",
    "NestedLocationResolver",
    "NestedLocationTransformer",
]


from dataclasses import dataclass
from typing import Iterable, Iterator, List, Optional, Tuple, Union

from beet import Context, Generator
from beet.core.utils import required_field
from tokenstream import InvalidSyntax, TokenStream, set_location

from mecha import (
    AlternativeParser,
    AstCommand,
    AstNode,
    AstResourceLocation,
    AstRoot,
    CommandSpec,
    CommentDisambiguation,
    CompilationDatabase,
    Dispatcher,
    Mecha,
    MutatingReducer,
    delegate,
    rule,
)
from mecha.contrib.relative_location import resolve_relative_location


def beet_default(ctx: Context):
    ctx.require(nested_location)


def nested_location(ctx: Context):
    ctx.require("mecha.contrib.nesting")

    mc = ctx.inject(Mecha)

    mc.spec.parsers["nested_location"] = parse_nested_location
    mc.spec.parsers["resource_location_or_tag"] = CommentDisambiguation(
        AlternativeParser(
            [
                delegate("nested_location"),
                mc.spec.parsers["resource_location_or_tag"],
            ]
        )
    )

    mc.transform.extend(
        NestedLocationTransformer(
            nested_location_resolver=ctx.inject(NestedLocationResolver)
        )
    )


@dataclass(frozen=True, slots=True)
class AstNestedLocation(AstResourceLocation):
    """Ast nested location node."""

    def get_value(self) -> str:
        raise set_location(UnresolvedNestedLocation(self), self)

    def get_canonical_value(self) -> str:
        raise set_location(UnresolvedNestedLocation(self), self)


class UnresolvedNestedLocation(InvalidSyntax):
    node: AstNestedLocation

    def __init__(self, node: AstNestedLocation):
        super().__init__(node)
        self.node = node

    def __str__(self) -> str:
        tag = "#" * self.node.is_tag
        return f'Unresolved nested location "{tag}~/{self.node.path}".'


def parse_nested_location(stream: TokenStream) -> AstNestedLocation:
    with stream.syntax(nested_location=r"#?~/[0-9a-z_./-]*"):
        token = stream.expect("nested_location")
        value = token.value

        if is_tag := value.startswith("#"):
            value = value[1:]

        path = value[2:]

        node = AstNestedLocation(is_tag=is_tag, path=path)
        return set_location(node, token)


class NestedLocationResolver:
    spec: CommandSpec
    steps: List[Dispatcher[AstRoot]]
    database: CompilationDatabase
    generate: Generator

    def __init__(self, ctx: Context):
        mc = ctx.inject(Mecha)
        self.spec = mc.spec
        self.steps = mc.steps
        self.database = mc.database
        self.generate = ctx.generate

    def resolve(
        self,
        nested_location: AstNestedLocation = AstNestedLocation(path=""),
    ) -> Tuple[str, str]:
        root, relative_path = self.concat_nested_path(
            nested_location,
            self.walk_location_hierarchy(
                self.spec,
                (
                    (command.identifier, command.arguments)
                    for command in reversed(self.steps[self.database.step].stack)
                    if (
                        isinstance(command, AstCommand)
                        and command.arguments
                        and isinstance(command.arguments[-1], (AstCommand, AstRoot))
                        and all(nested_location is not arg for arg in command.arguments)
                    )
                ),
            ),
        )

        if not root:
            current_file = self.database.current
            root = self.database[current_file].resource_location

        if not root:
            return resolve_relative_location(relative_path, self.generate.path("dummy"))

        return resolve_relative_location(relative_path, root, include_root_file=True)

    @staticmethod
    def walk_location_hierarchy(
        spec: CommandSpec,
        stack: Iterable[Tuple[str, Tuple[AstNode, ...]]],
    ) -> Iterator[AstResourceLocation]:
        for identifier, arguments in stack:
            prototype = spec.prototypes[identifier]

            for i, argument_node in enumerate(arguments):
                node = spec.tree.get(prototype.get_argument(i).scope)
                if (
                    node
                    and node.parser == "minecraft:function"
                    and (tail := spec.tree.get(*identifier.split(":")))
                    and tail.redirect != ("execute",)
                    and isinstance(argument_node, AstResourceLocation)
                ):
                    yield argument_node
                    break

    @staticmethod
    def concat_nested_path(
        *args: Union[
            Iterable[Union[AstResourceLocation, str]], AstResourceLocation, str
        ],
    ) -> Tuple[Optional[str], str]:
        root = None
        fragments: List[str] = []

        for segments in args:
            if isinstance(segments, (AstResourceLocation, str)):
                segments = [segments]
            for segment in segments:
                if isinstance(segment, AstNestedLocation):
                    value = segment.path
                elif isinstance(segment, AstResourceLocation):
                    value = segment.get_canonical_value()
                else:
                    value = segment
                if ":" in value:
                    root = value
                    break
                fragments.append(value)

        return root, "/".join(reversed(fragments))


@dataclass
class NestedLocationTransformer(MutatingReducer):
    nested_location_resolver: NestedLocationResolver = required_field()

    @rule(AstNestedLocation)
    def nested_location(self, node: AstNestedLocation):
        namespace, resolved = self.nested_location_resolver.resolve(node)
        resource_location = AstResourceLocation(
            is_tag=node.is_tag,
            namespace=namespace,
            path=resolved,
        )
        return set_location(resource_location, node)



"""Plugin that emits source mapping information."""

from pydantic import ConfigDict

__all__ = [
    "AstSourceMap",
    "SourceMapOptions",
    "SourceMapTransformer",
    "SourceMapSerializer",
    "source_map",
]


from dataclasses import dataclass, replace
from typing import List

from beet import Context, configurable
from beet.core.utils import required_field
from jinja2 import Template
from pydantic import BaseModel

from mecha import (
    AstChildren,
    AstCommandSentinel,
    AstRoot,
    CompilationDatabase,
    Mecha,
    MutatingReducer,
    Visitor,
    rule,
)


class SourceMapOptions(BaseModel):
    header: str = "# [source_map] {{ compilation_unit.filename or compilation_unit.resource_location }}"
    model_config = ConfigDict(extra="forbid")


def beet_default(ctx: Context):
    ctx.require(source_map)


@configurable(validator=SourceMapOptions)
def source_map(ctx: Context, opts: SourceMapOptions):
    mc = ctx.inject(Mecha)
    mc.transform.extend(
        SourceMapTransformer(
            database=mc.database,
            header_template=ctx.template.compile(opts.header),
        )
    )
    mc.serialize.extend(SourceMapSerializer())


@dataclass(frozen=True, slots=True)
class AstSourceMap(AstCommandSentinel):
    """Ast source map node."""

    header: str = required_field()

    compile_hints = {"skip_execute_inline_single_command": True}


@dataclass
class SourceMapTransformer(MutatingReducer):
    database: CompilationDatabase = required_field()
    header_template: Template = required_field()

    @rule(AstRoot)
    def insert_source_map(self, node: AstRoot):
        if not node.commands or not isinstance(node.commands[0], AstSourceMap):
            file_instance = self.database.current
            source_map = AstSourceMap(
                header=self.header_template.render(
                    file_instance=file_instance,
                    compilation_unit=self.database[file_instance],
                )
            )
            node = replace(node, commands=AstChildren([source_map, *node.commands]))
        return node


class SourceMapSerializer(Visitor):
    @rule(AstSourceMap)
    def source_map(self, node: AstSourceMap, result: List[str]):
        result.append(node.header)



"""Plugin for emitting ast."""

__all__ = [
    "DebugAstOptions",
    "DebugAstEmitter",
    "debug_ast",
]


from dataclasses import dataclass

from beet import Context, configurable
from beet.core.utils import required_field
from pydantic import BaseModel

from mecha import AstRoot, CompilationDatabase, Mecha, Visitor, rule


class DebugAstOptions(BaseModel):
    location: bool = False


def beet_default(ctx: Context):
    ctx.require(debug_ast)


@configurable(validator=DebugAstOptions)
def debug_ast(ctx: Context, opts: DebugAstOptions):
    mc = ctx.inject(Mecha)
    mc.steps[:] = [DebugAstEmitter(location=opts.location, database=mc.database)]


@dataclass
class DebugAstEmitter(Visitor):
    """Visitor that interrupts the compilation process and dumps the current ast."""

    location: bool = False
    database: CompilationDatabase = required_field()

    @rule(AstRoot)
    def debug_ast(self, node: AstRoot):
        exclude = None if self.location else {"location", "end_location"}
        self.database.current.text = node.dump(exclude=exclude) + "\n"
        return None



"""Plugin for parsing and resolving embeds in json and nbt strings."""

__all__ = [
    "AstJsonValueEmbed",
    "AstNbtValueEmbed",
    "EmbedHandler",
    "EmbedResolver",
    "EmbedParseCallback",
    "EmbedSerializeCallback",
]


from dataclasses import dataclass
from difflib import SequenceMatcher
from typing import Any, List, Optional, Protocol, TypeVar, Union

from beet import Context, TextFileBase
from beet.core.utils import required_field
from tokenstream import INITIAL_LOCATION, Preprocessor, SourceLocation, set_location

from mecha import (
    AstJsonValue,
    AstNbtValue,
    AstNode,
    CompilationDatabase,
    Diagnostic,
    DiagnosticError,
    Mecha,
    MutatingReducer,
    rule,
)

AstNodeType = TypeVar("AstNodeType", bound=AstNode)


@dataclass(frozen=True, slots=True)
class AstJsonValueEmbed(AstJsonValue):
    """Ast json value embed node."""

    embed: AstNode = required_field()


@dataclass(frozen=True, slots=True)
class AstNbtValueEmbed(AstNbtValue):
    """Ast nbt value embed node."""

    embed: AstNode = required_field()


def beet_default(ctx: Context):
    mc = ctx.inject(Mecha)
    embed_handler = ctx.inject(EmbedHandler)

    mc.steps.append(embed_handler.resolver)


class EmbedParseCallback(Protocol):
    """Callback required for parsing embed."""

    def __call__(
        self,
        source: TextFileBase[Any],
        *,
        using: str,
        preprocessor: Preprocessor,
    ) -> AstNode: ...


class EmbedSerializeCallback(Protocol):
    """Callback required for serializing embed."""

    def __call__(self, node: AstNode) -> str: ...


class EmbedHandler:
    """Service for handling embeds within json and nbt strings."""

    database: CompilationDatabase
    parse_callback: EmbedParseCallback
    serialize_callback: EmbedSerializeCallback
    resolver: "EmbedResolver"

    def __init__(self, arg: Union[Context, Mecha]):
        if isinstance(arg, Context):
            arg = arg.inject(Mecha)

        self.database = arg.database
        self.parse_callback = arg.parse
        self.serialize_callback = arg.serialize
        self.resolver = EmbedResolver(embed_handler=self)

    def parse(
        self,
        node: AstNodeType,
        source: Optional[TextFileBase[Any]] = None,
        *,
        using: str,
    ) -> AstNodeType:
        """Parse a string embed originating from the given source file.

        Defaults to the current source file.
        """
        if isinstance(node, (AstJsonValueEmbed, AstNbtValueEmbed)):
            return node
        if not isinstance(node, (AstJsonValue, AstNbtValue)):
            return node

        if source is None:
            source = self.database.current
        compilation_unit = self.database[source]

        if not isinstance(node.value, str):
            d = Diagnostic(
                "error",
                f'Couldn\'t parse value of type "{type(node.value).__name__}" as "{using}" embed.',
            )
            raise set_location(d, node)

        value = str(node.value)
        source_mappings: List[SourceLocation] = []
        preprocessed_mappings: List[SourceLocation] = []

        if (
            compilation_unit.source
            and not node.location.unknown
            and not node.end_location.unknown
        ):
            source_location = node.location
            preprocessed_location = INITIAL_LOCATION

            raw_value = compilation_unit.source[
                node.location.pos : node.end_location.pos
            ]

            index = 0
            raw_index = 0

            sequence_matcher = SequenceMatcher(None, value, raw_value)
            for i, j, size in sequence_matcher.get_matching_blocks():
                if size == 0:
                    continue

                source_mappings.append(source_location)
                preprocessed_mappings.append(preprocessed_location)

                extra = raw_value[raw_index:j]
                missing = value[index:i]

                if extra or missing:
                    source_location = source_location.skip_over(extra)
                    preprocessed_location = preprocessed_location.skip_over(missing)
                    source_mappings.append(source_location)
                    preprocessed_mappings.append(preprocessed_location)

                index = i + size
                raw_index = j + size

                source_location = source_location.skip_over(raw_value[j:raw_index])
                preprocessed_location = preprocessed_location.skip_over(value[i:index])

            source_mappings.append(source_location)
            preprocessed_mappings.append(preprocessed_location)

        try:
            embed = self.parse_callback(
                source,
                using=using,
                preprocessor=lambda _: (value, source_mappings, preprocessed_mappings),
            )
        except DiagnosticError as exc:
            raise exc.diagnostics.exceptions[0] from None

        if isinstance(node, AstJsonValue):
            embed_type = AstJsonValueEmbed
        else:
            embed_type = AstNbtValueEmbed

        return set_location(embed_type(value=node.value, embed=embed), node)  # pyright: ignore[reportReturnType]


@dataclass
class EmbedResolver(MutatingReducer):
    """Mutating reducer that serializes embeds back to regular json and nbt strings."""

    embed_handler: EmbedHandler = required_field()

    @rule(AstJsonValueEmbed)
    def json_embed(self, node: AstJsonValueEmbed):
        return set_location(
            AstJsonValue.from_value(self.embed_handler.serialize_callback(node.embed)),
            node,
        )

    @rule(AstNbtValueEmbed)
    def nbt_embed(self, node: AstNbtValueEmbed):
        return set_location(
            AstNbtValue.from_value(self.embed_handler.serialize_callback(node.embed)),
            node,
        )



from pathlib import Path

import click
from beet import ErrorMessage, Project
from beet.toolchain.cli import beet

from mecha import AstCacheBackend

pass_project = click.make_pass_decorator(Project)


@beet.command()
@pass_project
@click.argument("filename", type=click.Path(exists=True, dir_okay=False))
def ast(project: Project, filename: str):
    """Inspect cached mecha ast."""
    source_path = Path(filename).resolve()
    ast_path = project.cache["mecha"].get_path(f"{source_path}-ast")

    try:
        with ast_path.open("rb") as f:
            click.echo(AstCacheBackend().load(f).dump())
    except Exception as exc:
        raise ErrorMessage(f'No cached ast for "{filename}".') from exc



"""Plugin for compiling json files with mecha."""

__all__ = [
    "JsonFileCompilation",
    "AstJsonRoot",
    "AstJsonContent",
    "AstMergeJsonContent",
    "AstAppendJsonContent",
    "AstPrependJsonContent",
    "JsonRootParser",
    "JsonContentParser",
    "JsonFileHandler",
]


from dataclasses import dataclass
from typing import Any, Set, Type, Union

from beet import Context, DataPack, JsonFileBase, ResourcePack
from beet.core.utils import required_field
from tokenstream import TokenStream, set_location

from mecha import (
    AstChildren,
    AstCommand,
    AstCommandSentinel,
    AstJson,
    AstRoot,
    CompilationDatabase,
    Diagnostic,
    Dispatcher,
    FileTypeCompilationUnitProvider,
    Mecha,
    Parser,
    Visitor,
    delegate,
    rule,
)


@dataclass(frozen=True, slots=True)
class AstJsonRoot(AstRoot):
    """Ast json root node."""


@dataclass(frozen=True, slots=True)
class AstJsonContent(AstCommandSentinel):
    """Ast json content node."""

    @classmethod
    def create_root_node(cls, value: AstJson):
        content_node = cls(arguments=AstChildren([value]))
        content_node = set_location(content_node, value)
        root_node = AstJsonRoot(commands=AstChildren([content_node]))
        root_node = set_location(root_node, content_node)
        return root_node


@dataclass(frozen=True, slots=True)
class AstMergeJsonContent(AstJsonContent):
    """Ast merge json content node."""


@dataclass(frozen=True, slots=True)
class AstAppendJsonContent(AstJsonContent):
    """Ast append json content node."""


@dataclass(frozen=True, slots=True)
class AstPrependJsonContent(AstJsonContent):
    """Ast prepend json content node."""


def beet_default(ctx: Context):
    ctx.inject(JsonFileCompilation).activate()


class JsonFileCompilation:
    """Service for managing json file compilation."""

    mc: Mecha
    handler: Dispatcher[AstRoot]
    file_types: Set[Type[JsonFileBase[Any]]]
    active: bool

    def __init__(
        self,
        arg: Union[Context, Mecha],
        *packs: Union[ResourcePack, DataPack],
    ):
        if isinstance(arg, Context):
            packs += arg.packs
            self.mc = arg.inject(Mecha)
        else:
            self.mc = arg

        self.handler = JsonFileHandler(database=self.mc.database)
        self.mc.steps.append(self.handler)

        self.file_types = {  # pyright: ignore[reportAttributeAccessIssue]
            file_type
            for pack in packs
            for file_type in pack.get_file_types(extend=JsonFileBase)
        }

        self.active = False

    def activate(self):
        if self.active:
            return
        self.active = True

        self.mc.spec.parsers["root"] = JsonRootParser(
            database=self.mc.database,
            json_file_compilation=self,
            root_parser=self.mc.spec.parsers["root"],
        )

        self.mc.spec.parsers["root_item"] = JsonContentParser(
            root_item_parser=self.mc.spec.parsers["root_item"],
            json_parser=delegate("json"),
        )

        self.mc.providers.append(
            FileTypeCompilationUnitProvider(
                sorted(self.file_types, key=lambda t: t.snake_name),  # pyright: ignore[reportArgumentType]
                no_index=True,
            )
        )


@dataclass
class JsonRootParser:
    """Parser for json root."""

    database: CompilationDatabase
    json_file_compilation: JsonFileCompilation
    root_parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if "json_content" not in stream.data:
            json_content = (
                type(self.database.current) in self.json_file_compilation.file_types
            )
            with stream.provide(json_content=json_content):
                node = self.root_parser(stream)
            if json_content and isinstance(node, AstRoot):
                json_root = AstJsonRoot(commands=node.commands)
                node = set_location(json_root, node)
            return node
        return self.root_parser(stream)


@dataclass
class JsonContentParser:
    """Parser for json content."""

    root_item_parser: Parser
    json_parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if stream.data.get("json_content"):
            with stream.syntax(json=r"\{|\["):
                hint = stream.peek()
            if hint and hint.match("json"):
                with stream.ignore("newline"):
                    node = self.json_parser(stream)
                root_item = AstJsonContent(arguments=AstChildren([node]))
                return set_location(root_item, node)
        return self.root_item_parser(stream)


@dataclass
class JsonFileHandler(Visitor):
    """Handler for json files."""

    database: CompilationDatabase = required_field()

    @rule(AstRoot)
    def root(self, node: AstRoot):
        return node

    @rule(AstJsonRoot)
    def json_root(self, node: AstJsonRoot):
        for command in node.commands:
            yield command
        return None

    @rule(AstCommand)
    def command(self, node: AstCommand):
        d = Diagnostic(
            "error",
            f'Unexpected "{node.identifier.split(":")[0]}" command in json file.',
        )
        return set_location(d, node)

    @rule(AstJsonContent)
    def json_content(self, node: AstJsonContent):
        if node.arguments and isinstance(content := node.arguments[0], AstJson):
            target = self.database.current
            if isinstance(target, JsonFileBase):
                target_type = type(target)
                file_instance = target_type(content.evaluate())

                if isinstance(node, AstMergeJsonContent):
                    compilation_unit = self.database[target]
                    if compilation_unit.pack and compilation_unit.resource_location:
                        compilation_unit.pack[target_type].merge(  # pyright: ignore[reportCallIssue, reportArgumentType]
                            {compilation_unit.resource_location: file_instance}
                        )
                    elif not target.merge(file_instance):
                        target.data = file_instance.data

                elif isinstance(node, AstAppendJsonContent):
                    target.append(file_instance)  # pyright: ignore[reportAttributeAccessIssue]

                elif isinstance(node, AstPrependJsonContent):
                    target.prepend(file_instance)  # pyright: ignore[reportAttributeAccessIssue]

                else:
                    target.data = file_instance.data



__all__ = [
    "get_default_parsers",
    "get_parsers",
    "get_stream_spec",
    "get_stream_scope",
    "get_stream_properties",
    "get_stream_multiline",
    "get_stream_line_indentation",
    "UnrecognizedParser",
    "delegate",
    "consume_line_continuation",
    "parse_root",
    "parse_command",
    "parse_argument",
    "MultilineParser",
    "parse_bool",
    "NumericParser",
    "CoordinateParser",
    "IntegerConstraint",
    "MinMaxConstraint",
    "RestrictCoordinateConstraint",
    "StringParser",
    "parse_string_argument",
    "Vector2Parser",
    "Vector3Parser",
    "TypeConstraint",
    "JsonParser",
    "JsonObjectParser",
    "NbtParser",
    "NbtCompoundParser",
    "AdjacentConstraint",
    "ResourceLocationParser",
    "NoTagConstraint",
    "BlockParser",
    "BracketedListParser",
    "KeyValuePairParser",
    "ItemParser",
    "RemovedDefaultItemComponentParser",
    "ItemPredicateAlternativesParser",
    "NoDataTagsConstraint",
    "BasicLiteralParser",
    "RangeParser",
    "IntegerRangeConstraint",
    "LengthConstraint",
    "CommentDisambiguation",
    "TimeParser",
    "parse_uuid",
    "SelectorParser",
    "SelectorArgumentParser",
    "SelectorArgumentInvertConstraint",
    "SelectorArgumentNoValueConstraint",
    "parse_selector_scores",
    "parse_selector_advancements",
    "SelectorPlayerConstraint",
    "SelectorTypeConstraint",
    "SelectorSingleConstraint",
    "SelectorAmountConstraint",
    "EntityParser",
    "parse_wildcard",
    "parse_message",
    "NbtPathParser",
    "ParticleParser",
    "AggregateParser",
    "SingleLineConstraint",
    "AlternativeParser",
    "NUMBER_PATTERN",
    "JSON_STRING_PATTERN",
    "NBT_QUOTED_STRING_PATTERN",
]


from contextlib import nullcontext
from dataclasses import dataclass, field
from functools import partial
from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    overload,
)
from uuid import UUID

from beet import LATEST_MINECRAFT_VERSION
from beet.core.utils import VersionNumber, split_version
from nbtlib import Byte, Double, Float, Int, Long, OutOfRange, Short, String
from tokenstream import InvalidSyntax, SourceLocation, TokenStream, set_location

from .ast import (
    AstAdvancementPredicate,
    AstBlock,
    AstBlockMarkerParticleParameters,
    AstBlockParticleParameters,
    AstBlockState,
    AstBool,
    AstChildren,
    AstColor,
    AstColorReset,
    AstCommand,
    AstCoordinate,
    AstDustColorTransitionParticleParameters,
    AstDustParticleParameters,
    AstEntityAnchor,
    AstFallingDustParticleParameters,
    AstGamemode,
    AstGreedy,
    AstHeightmap,
    AstHexColor,
    AstItemComponent,
    AstItemParticleParameters,
    AstItemPredicate,
    AstItemPredicateAlternatives,
    AstItemPredicateTestComponent,
    AstItemPredicateTestPredicate,
    AstItemRemovedDefaultComponent,
    AstItemSlot,
    AstItemSlots,
    AstItemStack,
    AstJson,
    AstJsonArray,
    AstJsonObject,
    AstJsonObjectEntry,
    AstJsonObjectKey,
    AstJsonValue,
    AstLegacyScoreboardSlot,
    AstLiteral,
    AstMacroLine,
    AstMacroLineText,
    AstMacroLineVariable,
    AstMessage,
    AstMessageText,
    AstNbt,
    AstNbtBool,
    AstNbtByteArray,
    AstNbtCompound,
    AstNbtCompoundEntry,
    AstNbtCompoundKey,
    AstNbtIntArray,
    AstNbtList,
    AstNbtLongArray,
    AstNbtPath,
    AstNbtPathKey,
    AstNbtPathSubscript,
    AstNbtValue,
    AstNode,
    AstNumber,
    AstObjective,
    AstObjectiveCriteria,
    AstParticle,
    AstPlayerName,
    AstRange,
    AstResourceLocation,
    AstRoot,
    AstScoreboardOperation,
    AstScoreboardSlot,
    AstSculkChargeParticleParameters,
    AstSelector,
    AstSelectorAdvancementMatch,
    AstSelectorAdvancementPredicateMatch,
    AstSelectorAdvancements,
    AstSelectorArgument,
    AstSelectorScoreMatch,
    AstSelectorScores,
    AstShriekParticleParameters,
    AstSortOrder,
    AstString,
    AstSwizzle,
    AstTeam,
    AstTemplateMirror,
    AstTemplateRotation,
    AstTime,
    AstUUID,
    AstVector2,
    AstVector3,
    AstVibrationParticleParameters,
    AstWildcard,
    AstWord,
)
from .config import CommandTree
from .error import MechaError
from .spec import CommandSpec, Parser
from .utils import JsonQuoteHelper, NbtQuoteHelper, QuoteHelper, string_to_number

NUMBER_PATTERN: str = r"-?(?:\d+\.?\d*|\.\d+)"

JSON_STRING_PATTERN: str = r'"(?:\\.|[^\\\n])*?"'
NBT_QUOTED_STRING_PATTERN: str = r'"(?:\\.|[^\\\n])*?"' "|" r"'(?:\\.|[^\\\n])*?'"


def get_default_parsers() -> Dict[str, Parser]:
    """Return the default parsers."""
    json_parser = JsonParser()
    json_parser.object_entry_parser = delegate("json_object_entry")
    json_parser.array_element_parser = delegate("json_array_element")
    json_parser.recursive_parser = delegate("json")

    nbt_parser = NbtParser()
    nbt_parser.compound_entry_parser = delegate("nbt_compound_entry")
    nbt_parser.list_or_array_element_parser = delegate("nbt_list_or_array_element")
    nbt_parser.recursive_parser = delegate("nbt")

    return {
        ################################################################################
        # Primitives
        ################################################################################
        "bool": parse_bool,
        "numeric": NumericParser(),
        "integer": IntegerConstraint(delegate("numeric")),
        "coordinate": CoordinateParser(),
        "integer_coordinate": IntegerConstraint(delegate("coordinate")),
        "time": TimeParser(),
        "word": StringParser(type="word"),
        "phrase": StringParser(type="phrase"),
        "greedy": StringParser(type="greedy"),
        "json": json_parser,
        "json_object_entry": json_parser.parse_object_entry,
        "json_array_element": delegate("json"),
        "json_object": JsonObjectParser(delegate("json")),
        "nbt": nbt_parser,
        "nbt_compound_entry": nbt_parser.parse_compound_entry,
        "nbt_list_or_array_element": delegate("nbt"),
        "nbt_compound": NbtCompoundParser(delegate("nbt")),
        "adjacent_nbt_compound": AdjacentConstraint(delegate("nbt_compound"), r"\{"),
        "nbt_path": NbtPathParser(
            integer_parser=delegate("integer"),
            nbt_compound_parser=delegate("nbt_compound"),
        ),
        "range": RangeParser(),
        "integer_range": IntegerRangeConstraint(delegate("range")),
        "resource_location_or_tag": CommentDisambiguation(ResourceLocationParser()),
        "resource_location": NoTagConstraint(delegate("resource_location_or_tag")),
        "resource_location_or_nbt": AlternativeParser(
            [delegate("resource_location"), delegate("nbt_compound")]
        ),
        "uuid": parse_uuid,
        "objective": BasicLiteralParser(AstObjective),
        "objective_criteria": BasicLiteralParser(AstObjectiveCriteria),
        "scoreboard_operation": BasicLiteralParser(AstScoreboardOperation),
        "player_name": CommentDisambiguation(BasicLiteralParser(AstPlayerName)),
        "scoreboard_slot": BasicLiteralParser(AstScoreboardSlot),
        "swizzle": BasicLiteralParser(AstSwizzle),
        "team": BasicLiteralParser(AstTeam),
        "advancement_predicate": BasicLiteralParser(AstAdvancementPredicate),
        "wildcard": parse_wildcard,
        "color": BasicLiteralParser(AstColor),
        "color_reset": BasicLiteralParser(AstColorReset),
        "hex_color": HexColorParser(),
        "sort_order": BasicLiteralParser(AstSortOrder),
        "gamemode": BasicLiteralParser(AstGamemode),
        "heightmap": BasicLiteralParser(AstHeightmap),
        "entity_anchor": BasicLiteralParser(AstEntityAnchor),
        "template_rotation": BasicLiteralParser(AstTemplateRotation),
        "template_mirror": BasicLiteralParser(AstTemplateMirror),
        "block_predicate": BlockParser(
            resource_location_parser=delegate("resource_location_or_tag"),
            block_states_parser=AdjacentConstraint(
                parser=MultilineParser(
                    BracketedListParser(
                        KeyValuePairParser(
                            node_type=AstBlockState,
                            key_parser=StringParser(type="phrase"),
                            value_parser=delegate("phrase"),
                        )
                    )
                ),
                hint=r"\[",
            ),
            data_tags_parser=delegate("adjacent_nbt_compound"),
        ),
        "block_state": BlockParser(
            resource_location_parser=delegate("resource_location"),
            block_states_parser=AdjacentConstraint(
                parser=MultilineParser(
                    BracketedListParser(
                        KeyValuePairParser(
                            node_type=AstBlockState,
                            key_parser=StringParser(type="phrase"),
                            value_parser=delegate("phrase"),
                        )
                    )
                ),
                hint=r"\[",
            ),
            data_tags_parser=delegate("adjacent_nbt_compound"),
        ),
        "item_predicate": ItemParser(
            node_type=AstItemPredicate,
            identifier_parser=AlternativeParser(
                [delegate("resource_location_or_tag"), delegate("wildcard")]
            ),
            arguments_parser=AdjacentConstraint(
                parser=MultilineParser(
                    BracketedListParser(
                        ItemPredicateAlternativesParser(
                            key_parser=delegate("resource_location"),
                            value_parser=delegate("nbt"),
                        )
                    )
                ),
                hint=r"\[",
            ),
            data_tags_parser=delegate("adjacent_nbt_compound"),
        ),
        "item_slot": BasicLiteralParser(AstItemSlot),
        "item_slots": BasicLiteralParser(AstItemSlots),
        "item_stack": ItemParser(
            node_type=AstItemStack,
            identifier_parser=delegate("resource_location"),
            arguments_parser=AdjacentConstraint(
                parser=MultilineParser(
                    BracketedListParser(
                        AlternativeParser(
                            [
                                RemovedDefaultItemComponentParser(
                                    delegate("resource_location")
                                ),
                                KeyValuePairParser(
                                    node_type=AstItemComponent,
                                    key_parser=delegate("resource_location"),
                                    value_parser=delegate("nbt"),
                                ),
                            ]
                        )
                    )
                ),
                hint=r"\[",
            ),
            data_tags_parser=delegate("adjacent_nbt_compound"),
        ),
        "message": parse_message,
        "block_pos": Vector3Parser(coordinate_parser=delegate("coordinate")),
        "column_pos": Vector2Parser(
            coordinate_parser=RestrictCoordinateConstraint(
                parser=delegate("integer_coordinate"),
                disallow=["local"],
            ),
        ),
        "rotation": Vector2Parser(
            coordinate_parser=RestrictCoordinateConstraint(
                parser=delegate("coordinate"),
                disallow=["local"],
            ),
        ),
        "vec2": Vector2Parser(coordinate_parser=delegate("coordinate")),
        "vec3": Vector3Parser(coordinate_parser=delegate("coordinate")),
        "entity": EntityParser(
            selector_parser=SelectorTypeConstraint(
                SelectorAmountConstraint(delegate("selector"))
            ),
        ),
        "score_holder": AlternativeParser(
            [
                EntityParser(SelectorAmountConstraint(delegate("selector"))),
                delegate("wildcard"),
            ]
        ),
        "game_profile": EntityParser(
            selector_parser=SelectorPlayerConstraint(delegate("selector")),
        ),
        ################################################################################
        # Particle
        ################################################################################
        "particle": ParticleParser(
            resource_location_parser=delegate("resource_location"),
            generic_parameters_parser=MultilineParser(delegate("nbt_compound")),
        ),
        "particle:minecraft:dust": AggregateParser(
            type=AstDustParticleParameters,
            fields={
                "red": delegate("numeric"),
                "green": delegate("numeric"),
                "blue": delegate("numeric"),
                "size": delegate("numeric"),
            },
        ),
        "particle:minecraft:dust_color_transition": AggregateParser(
            type=AstDustColorTransitionParticleParameters,
            fields={
                "red": delegate("numeric"),
                "green": delegate("numeric"),
                "blue": delegate("numeric"),
                "size": delegate("numeric"),
                "end_red": delegate("numeric"),
                "end_green": delegate("numeric"),
                "end_blue": delegate("numeric"),
            },
        ),
        "particle:minecraft:block": AggregateParser(
            type=AstBlockParticleParameters,
            fields={"block": delegate("block_state")},
        ),
        "particle:minecraft:falling_dust": AggregateParser(
            type=AstFallingDustParticleParameters,
            fields={"block": delegate("block_state")},
        ),
        "particle:minecraft:item": AggregateParser(
            type=AstItemParticleParameters,
            fields={"item": delegate("item_stack")},
        ),
        "particle:minecraft:vibration": AggregateParser(
            type=AstVibrationParticleParameters,
            fields={
                "destination_x": delegate("numeric"),
                "destination_y": delegate("numeric"),
                "destination_z": delegate("numeric"),
                "duration": delegate("integer"),
            },
        ),
        "particle:minecraft:block_marker": AggregateParser(
            type=AstBlockMarkerParticleParameters,
            fields={"block": delegate("block_state")},
        ),
        "particle:minecraft:sculk_charge": AggregateParser(
            type=AstSculkChargeParticleParameters,
            fields={"angle": delegate("numeric")},
        ),
        "particle:minecraft:shriek": AggregateParser(
            type=AstShriekParticleParameters,
            fields={"delay": delegate("integer")},
        ),
        ################################################################################
        # Selector
        ################################################################################
        "selector": MultilineParser(SelectorParser()),
        "selector:argument": SelectorArgumentInvertConstraint(
            SelectorArgumentNoValueConstraint(
                SelectorArgumentParser(key_parser=StringParser(type="phrase")),
                allow_no_value=["tag", "team"],
            ),
            allow_invert=[
                "tag",
                "team",
                "gamemode",
                "name",
                "type",
                "nbt",
                "predicate",
            ],
        ),
        "selector:argument:x": delegate("numeric"),
        "selector:argument:y": delegate("numeric"),
        "selector:argument:z": delegate("numeric"),
        "selector:argument:distance": delegate("range"),
        "selector:argument:dx": delegate("numeric"),
        "selector:argument:dy": delegate("numeric"),
        "selector:argument:dz": delegate("numeric"),
        "selector:argument:scores": parse_selector_scores,
        "selector:argument:tag": delegate("word"),
        "selector:argument:team": delegate("team"),
        "selector:argument:limit": delegate("integer"),
        "selector:argument:sort": delegate("sort_order"),
        "selector:argument:level": delegate("integer_range"),
        "selector:argument:gamemode": delegate("gamemode"),
        "selector:argument:name": delegate("phrase"),
        "selector:argument:x_rotation": delegate("range"),
        "selector:argument:y_rotation": delegate("range"),
        "selector:argument:type": delegate("resource_location_or_tag"),
        "selector:argument:nbt": delegate("nbt_compound"),
        "selector:argument:advancements": parse_selector_advancements,
        "selector:argument:predicate": delegate("resource_location"),
        ################################################################################
        # Command
        ################################################################################
        "root": parse_root,
        "root_item": AlternativeParser([delegate("macro_line"), delegate("command")]),
        "command": parse_command,
        "command:argument": parse_argument,
        "command:argument:brigadier:bool": delegate("bool"),
        "command:argument:brigadier:double": MinMaxConstraint(delegate("numeric")),
        "command:argument:brigadier:float": MinMaxConstraint(delegate("numeric")),
        "command:argument:brigadier:integer": MinMaxConstraint(delegate("integer")),
        "command:argument:brigadier:long": MinMaxConstraint(delegate("integer")),
        "command:argument:brigadier:string": parse_string_argument,
        "command:argument:minecraft:angle": RestrictCoordinateConstraint(
            parser=delegate("coordinate"),
            disallow=["local"],
        ),
        "command:argument:minecraft:block_pos": delegate("block_pos"),
        "command:argument:minecraft:block_predicate": MultilineParser(
            delegate("block_predicate")
        ),
        "command:argument:minecraft:block_state": MultilineParser(
            delegate("block_state")
        ),
        "command:argument:minecraft:color": AlternativeParser(
            [
                delegate("color"),
                delegate("color_reset"),
            ]
        ),
        "command:argument:minecraft:hex_color": delegate("hex_color"),
        "command:argument:minecraft:column_pos": delegate("column_pos"),
        "command:argument:minecraft:component": MultilineParser(delegate("nbt")),
        "command:argument:minecraft:style": MultilineParser(delegate("nbt")),
        "command:argument:minecraft:dimension": delegate("resource_location"),
        "command:argument:minecraft:dialog": MultilineParser(
            delegate("resource_location_or_nbt")
        ),
        "command:argument:minecraft:entity": delegate("entity"),
        "command:argument:minecraft:entity_anchor": delegate("entity_anchor"),
        "command:argument:minecraft:entity_summon": delegate("resource_location"),
        "command:argument:minecraft:template_rotation": delegate("template_rotation"),
        "command:argument:minecraft:template_mirror": delegate("template_mirror"),
        "command:argument:minecraft:float_range": delegate("range"),
        "command:argument:minecraft:function": delegate("resource_location_or_tag"),
        "command:argument:minecraft:loot_table": MultilineParser(
            delegate("resource_location_or_nbt")
        ),
        "command:argument:minecraft:loot_predicate": MultilineParser(
            delegate("resource_location_or_nbt")
        ),
        "command:argument:minecraft:loot_modifier": MultilineParser(
            delegate("resource_location_or_nbt")
        ),
        "command:argument:minecraft:game_profile": delegate("game_profile"),
        "command:argument:minecraft:gamemode": delegate("gamemode"),
        "command:argument:minecraft:heightmap": delegate("heightmap"),
        "command:argument:minecraft:int_range": delegate("integer_range"),
        "command:argument:minecraft:item_enchantment": delegate("word"),
        "command:argument:minecraft:item_predicate": MultilineParser(
            delegate("item_predicate")
        ),
        "command:argument:minecraft:item_slot": delegate("item_slot"),
        "command:argument:minecraft:item_slots": delegate("item_slots"),
        "command:argument:minecraft:item_stack": MultilineParser(
            delegate("item_stack")
        ),
        "command:argument:minecraft:message": delegate("message"),
        "command:argument:minecraft:mob_effect": delegate("resource_location"),
        "command:argument:minecraft:nbt_compound_tag": MultilineParser(
            delegate("nbt_compound")
        ),
        "command:argument:minecraft:nbt_path": MultilineParser(delegate("nbt_path")),
        "command:argument:minecraft:nbt_tag": MultilineParser(delegate("nbt")),
        "command:argument:minecraft:objective": AlternativeParser(
            [
                delegate("objective"),
                delegate("wildcard"),
            ]
        ),
        "command:argument:minecraft:objective_criteria": delegate("objective_criteria"),
        "command:argument:minecraft:operation": delegate("scoreboard_operation"),
        "command:argument:minecraft:particle": delegate("particle"),
        "command:argument:minecraft:resource_location": delegate("resource_location"),
        "command:argument:minecraft:resource": delegate("resource_location"),
        "command:argument:minecraft:resource_key": delegate("resource_location"),
        "command:argument:minecraft:resource_or_tag": delegate(
            "resource_location_or_tag"
        ),
        "command:argument:minecraft:resource_or_tag_key": delegate(
            "resource_location_or_tag"
        ),
        # TODO: support wildcards
        "command:argument:minecraft:resource_selector": delegate("resource_location"),
        "command:argument:minecraft:rotation": delegate("rotation"),
        "command:argument:minecraft:score_holder": delegate("score_holder"),
        "command:argument:minecraft:scoreboard_slot": delegate("scoreboard_slot"),
        "command:argument:minecraft:swizzle": delegate("swizzle"),
        "command:argument:minecraft:team": delegate("team"),
        "command:argument:minecraft:time": delegate("time"),
        "command:argument:minecraft:uuid": delegate("uuid"),
        "command:argument:minecraft:vec2": delegate("vec2"),
        "command:argument:minecraft:vec3": delegate("vec3"),
        ################################################################################
        # Macro line
        ################################################################################
        "macro_line": parse_macro_line,
        "macro_line_text": BasicLiteralParser(AstMacroLineText),
        "macro_line_variable": BasicLiteralParser(AstMacroLineVariable),
    }


def get_parsers(version: VersionNumber = LATEST_MINECRAFT_VERSION) -> Dict[str, Parser]:
    """Return parsers for a specific version."""
    version = split_version(version)

    parsers = get_default_parsers()

    if version < (1, 18):
        parsers["objective"] = LengthConstraint(parsers["objective"], 16)
        parsers["player_name"] = LengthConstraint(parsers["player_name"], 40)

    if version < (1, 20):
        parsers["scoreboard_slot"] = BasicLiteralParser(AstLegacyScoreboardSlot)

    if version < (1, 21):
        parsers["command:argument:minecraft:component"] = MultilineParser(
            delegate("json")
        )
        parsers["command:argument:minecraft:style"] = MultilineParser(delegate("json"))

    return parsers


def get_stream_spec(stream: TokenStream) -> CommandSpec:
    """Return the command specification associated with the token stream."""
    return stream.data["spec"]


def get_stream_scope(stream: TokenStream) -> Tuple[str, ...]:
    """Return the current scope associated with the token stream."""
    return stream.data.get("scope", ())


def get_stream_properties(stream: TokenStream) -> Dict[str, Any]:
    """Return the current command node properties associated with the token stream."""
    return stream.data.get("properties", {})


def get_stream_multiline(stream: TokenStream) -> bool:
    """Return whether the token stream is currently parsing in multiline mode."""
    return stream.data.get("multiline", False)


def get_stream_line_indentation(stream: TokenStream) -> int:
    """Return the indentation level associated with the current line."""
    return stream.data.get("line_indentation", stream.indentation[-1])


class UnrecognizedParser(MechaError):
    """Raised when delegating to an unrecognized parser."""

    parser: str

    def __init__(self, parser: str):
        super().__init__(parser)
        self.parser = parser


@overload
def delegate(parser: str) -> Parser: ...


@overload
def delegate(parser: str, stream: TokenStream) -> Any: ...


def delegate(parser: str, stream: Optional[TokenStream] = None) -> Any:
    """Delegate parsing to a registered subparser."""
    if stream is None:
        return partial(delegate, parser)

    spec = get_stream_spec(stream)

    if parser not in spec.parsers:
        raise UnrecognizedParser(parser)

    return spec.parsers[parser](stream)


def consume_line_continuation(stream: TokenStream) -> bool:
    """Consume newlines and leading whitespace if there's a line continuation."""
    level = get_stream_line_indentation(stream)
    multiline = get_stream_multiline(stream)

    if not multiline:
        return False

    with stream.checkpoint() as commit:
        if list(stream.collect("newline")):
            with stream.intercept("whitespace"):
                if (
                    (whitespace := stream.get("whitespace"))
                    and not stream.get("eof")
                    and level < len(whitespace.value.expandtabs())
                ):
                    commit()
                    return True

    return False


def parse_root(stream: TokenStream) -> AstRoot:
    """Parse root."""
    start = stream.peek()

    if not start:
        node = AstRoot(commands=AstChildren[AstCommand]())
        return set_location(node, SourceLocation(0, 1, 1))

    commands: List[AstCommand] = []

    with stream.ignore("comment"):
        for _ in stream.peek_until():
            if stream.get("newline"):
                continue
            if stream.get("eof"):
                break
            commands.append(delegate("root_item", stream))

    node = AstRoot(commands=AstChildren(commands))
    return set_location(node, start, stream.current)


def parse_command(stream: TokenStream) -> AstCommand:
    """Parse command."""
    spec = get_stream_spec(stream)
    scope = get_stream_scope(stream)
    level = get_stream_line_indentation(stream)

    tree = spec.tree.get(scope)

    arguments: List[AstNode] = []
    reached_terminal = False

    location = None
    end_location = None

    while tree:
        with stream.checkpoint() as commit:
            literal = stream.expect("literal")

            if child := tree.get_literal(literal.value):
                if not child.children:
                    if stream.peek():
                        stream.expect("newline", "eof")
                    reached_terminal = True

                if location is None:
                    location = literal.location
                end_location = literal.end_location
                scope = scope + (literal.value,)
                tree = child

                commit()

        if commit.rollback:
            choices: List[Tuple[str, CommandTree]] = list((tree.children or {}).items())
            choices.sort(key=lambda p: p[1].type != "argument")

            if tree.executable:
                choices.append(("", tree))

            for (name, child), alternative in stream.choose(*choices):
                with (
                    alternative,
                    stream.provide(
                        scope=scope + (name,),
                        line_indentation=level,
                    ),
                ):
                    literal = None
                    argument = None

                    if tree is child and tree.executable:
                        if stream.peek():
                            stream.expect("newline", "eof")
                        reached_terminal = True
                        continue

                    elif child.type == "literal":
                        literal = stream.expect(("literal", name))
                        if location is None:
                            location = literal.location

                    elif child.type == "argument":
                        argument = delegate("command:argument", stream)
                        if location is None:
                            location = argument.location

                    if not child.children:
                        if stream.peek():
                            stream.expect("newline", "eof")
                        reached_terminal = True

                    if literal:
                        end_location = literal.end_location
                    elif argument:
                        arguments.append(argument)
                        end_location = argument.end_location

                    scope = stream.data["scope"]
                    tree = child

        if reached_terminal:
            break

        with stream.provide(line_indentation=level):
            consume_line_continuation(stream)

        target = spec.tree.get(tree.redirect)
        recursive = target and target.subcommand

        if tree.subcommand or recursive:
            if tree.executable and (not stream.peek() or stream.get("newline", "eof")):
                break
            subcommand_scope = tree.redirect if tree.redirect is not None else scope
            with (
                stream.alternative(bool(target and target.children)),
                stream.provide(
                    scope=subcommand_scope,
                    line_indentation=level,
                ),
            ):
                node = delegate("command", stream)
                arguments.append(node)
                end_location = node.end_location
                scope += ("subcommand",)
                break

    node = AstCommand(identifier=":".join(scope), arguments=AstChildren(arguments))
    return set_location(node, location, end_location)


def parse_argument(stream: TokenStream) -> AstNode:
    """Parse argument."""
    spec = get_stream_spec(stream)
    scope = get_stream_scope(stream)

    tree = spec.tree.get(scope)

    if tree and tree.parser:
        with stream.provide(properties=tree.properties or {}):
            return delegate(f"command:argument:{tree.parser}", stream)

    raise ValueError(f"Missing argument parser in command tree {scope}.")


@dataclass
class MultilineParser:
    """Allow a parser to span over multiple lines."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.checkpoint(), stream.intercept("newline"):
            newline = stream.get("newline")

        if not newline and get_stream_multiline(stream):
            with stream.ignore("newline"):
                return self.parser(stream)

        return self.parser(stream)


def parse_bool(stream: TokenStream) -> AstBool:
    """Parse bool."""
    with stream.syntax(literal=r"\w+"):
        token = stream.expect_any(("literal", "true"), ("literal", "false"))
    node = AstBool(value=token.value == "true")
    return set_location(node, token)


@dataclass
class NumericParser:
    """Parser for numeric values."""

    name: str = "number"
    pattern: str = NUMBER_PATTERN

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(**{self.name: self.pattern}):
            stream.expect(self.name)
        return self.create_node(stream)

    def create_node(self, stream: TokenStream) -> Any:
        """Create the ast node."""
        token = stream.current
        node = AstNumber(value=string_to_number(token.value))
        return set_location(node, token)


@dataclass
class HexColorParser:
    """Parser for hex color values."""

    name: str = "hex_color"
    pattern: str = r"[0-9a-fA-F]{6}|[0-9a-fA-F]{3}"

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(**{self.name: self.pattern}):
            stream.expect(self.name)
        return self.create_node(stream)

    def create_node(self, stream: TokenStream) -> Any:
        """Create the ast node."""
        token = stream.current
        node = AstHexColor(value=token.value)
        return set_location(node, token)


@dataclass
class CoordinateParser(NumericParser):
    """Parser for coordinates."""

    name: str = "coordinate"
    pattern: str = rf"(?:[~^]?{NUMBER_PATTERN}|[~^])(?!\w|-|\.)"

    def create_node(self, stream: TokenStream) -> Any:
        token = stream.current
        value = token.value

        if token.value.startswith("~"):
            coordinate_type = "relative"
            value = value[1:]
        elif token.value.startswith("^"):
            coordinate_type = "local"
            value = value[1:]
        else:
            coordinate_type = "absolute"

        if not value:
            value = "0"

        node = AstCoordinate(type=coordinate_type, value=string_to_number(value))
        return set_location(node, token)


@dataclass
class IntegerConstraint:
    """Constraint that disallows decimal numeric values."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), (AstNumber, AstCoordinate)):
            if not isinstance(node.value, int):
                raise node.emit_error(InvalidSyntax("Expected integer value."))
        return node


@dataclass
class MinMaxConstraint:
    """Constraint that checks that the value conforms to the min and max properties."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        properties = get_stream_properties(stream)

        if isinstance(node := self.parser(stream), AstNumber):
            if "min" in properties and node.value < properties["min"]:
                exc = InvalidSyntax(
                    f"Expected value to be at least {properties['min']}."
                )
                raise node.emit_error(exc)
            if "max" in properties and node.value > properties["max"]:
                exc = InvalidSyntax(
                    f"Expected value to be at most {properties['max']}."
                )
                raise node.emit_error(exc)

        return node


@dataclass
class RestrictCoordinateConstraint:
    """Constraint that disallows certain types of coordinates."""

    parser: Parser
    disallow: List[Literal["absolute", "relative", "local"]]

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstCoordinate):
            if node.type in self.disallow:
                exc = InvalidSyntax(f"Specifying {node.type} coordinates not allowed.")
                raise node.emit_error(exc)
        return node


@dataclass
class StringParser:
    """Parser for string values."""

    type: Literal["word", "phrase", "greedy"]
    quote_helper: QuoteHelper = field(default_factory=QuoteHelper)

    def __call__(self, stream: TokenStream) -> Any:
        if self.type == "greedy":
            with stream.intercept("whitespace"):
                while stream.get("whitespace"):
                    continue
            with stream.syntax(line=AstGreedy.regex.pattern):
                token = stream.expect("line")
            node = AstGreedy(value=token.value)
        else:
            with stream.syntax(
                word=AstWord.regex.pattern,
                quoted_string=r'"(?:\\.|[^\\\n])*?"' "|" r"'(?:\\.|[^\\\n])*?'",
            ):
                if self.type == "word":
                    token = stream.expect("word")
                    node = AstWord(value=token.value)
                else:
                    token = stream.expect_any("word", "quoted_string")
                    node = AstString(value=self.quote_helper.unquote_string(token))
        return set_location(node, token)


def parse_string_argument(stream: TokenStream) -> Any:
    """Parse string argument."""
    properties = get_stream_properties(stream)
    string_type = properties.get("type", "word")

    if string_type not in ["word", "phrase", "greedy"]:
        raise ValueError(f"Invalid string type {string_type!r}.")

    return delegate(string_type, stream)


@dataclass
class Vector2Parser:
    """Parser for vector2."""

    coordinate_parser: Parser

    def __call__(self, stream: TokenStream) -> AstVector2:
        x = self.coordinate_parser(stream)
        y = self.coordinate_parser(stream)

        node = AstVector2(x=x, y=y)
        return set_location(node, x, y)


@dataclass
class Vector3Parser:
    """Parser for vector3."""

    coordinate_parser: Parser

    def __call__(self, stream: TokenStream) -> AstVector3:
        x = self.coordinate_parser(stream)
        y = self.coordinate_parser(stream)
        z = self.coordinate_parser(stream)

        node = AstVector3(x=x, y=y, z=z)
        return set_location(node, x, z)


@dataclass
class TypeConstraint:
    """Constraint that only allows specific instances."""

    parser: Parser
    type: Union[Type[Any], Tuple[Type[Any], ...]]
    message: str

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        if not isinstance(node, self.type):
            raise node.emit_error(InvalidSyntax(self.message))

        return node


@dataclass
class JsonParser:
    """Parser for json values."""

    curly_pattern: str = r"\{|\}"
    bracket_pattern: str = r"\[|\]"
    string_pattern: str = JSON_STRING_PATTERN
    number_pattern: str = r"-?(?:0|[1-9]\d*)(?:\.\d+)?(?:[eE][+-]?\d+)?\b"
    colon_pattern: str = r":"
    comma_pattern: str = r","
    null_pattern: str = r"\bnull\b"
    true_pattern: str = r"\btrue\b"
    false_pattern: str = r"\bfalse\b"

    object_entry_parser: Parser = field(init=False)
    array_element_parser: Parser = field(init=False)
    recursive_parser: Parser = field(init=False)

    quote_helper: QuoteHelper = field(default_factory=JsonQuoteHelper)

    def __post_init__(self):
        self.object_entry_parser = self.parse_object_entry
        self.array_element_parser = self
        self.recursive_parser = self

    def __call__(self, stream: TokenStream) -> AstJson:
        with stream.syntax(
            curly=self.curly_pattern,
            bracket=self.bracket_pattern,
            string=self.string_pattern,
            number=self.number_pattern,
            colon=self.colon_pattern,
            comma=self.comma_pattern,
            null=self.null_pattern,
            true=self.true_pattern,
            false=self.false_pattern,
        ):
            curly, bracket, string, number, null, true, false = stream.expect(
                ("curly", "{"),
                ("bracket", "["),
                "string",
                "number",
                "null",
                "true",
                "false",
            )

            if curly:
                entries: List[AstJsonObjectEntry] = []

                for _ in stream.peek_until(("curly", "}")):
                    entries.append(self.object_entry_parser(stream))

                    if not stream.get("comma"):
                        stream.expect(("curly", "}"))
                        break

                node = AstJsonObject(entries=AstChildren(entries))
                return set_location(node, curly, stream.current)

            elif bracket:
                elements: List[AstJson] = []

                for _ in stream.peek_until(("bracket", "]")):
                    elements.append(self.array_element_parser(stream))

                    if not stream.get("comma"):
                        stream.expect(("bracket", "]"))
                        break

                node = AstJsonArray(elements=AstChildren(elements))
                return set_location(node, bracket, stream.current)

            if null:
                value = None
            elif true:
                value = True
            elif false:
                value = False
            elif string:
                value = self.quote_helper.unquote_string(string)
            elif number:
                value = string_to_number(number.value)

            node = AstJsonValue(value=value)  # pyright: ignore[reportPossiblyUnboundVariable]
            return set_location(node, stream.current)

    def parse_object_entry(self, stream: TokenStream) -> AstJsonObjectEntry:
        """Parse json object entry."""
        key = stream.expect("string")
        key_node = AstJsonObjectKey(value=self.quote_helper.unquote_string(key))
        key_node = set_location(key_node, key)

        stream.expect("colon")

        value_node = self.recursive_parser(stream)

        entry_node = AstJsonObjectEntry(key=key_node, value=value_node)
        return set_location(entry_node, key_node, value_node)


@dataclass
class JsonObjectParser(TypeConstraint):
    """Parser for json objects."""

    parser: Parser = field(default_factory=JsonParser)
    type: Union[Type[Any], Tuple[Type[Any], ...]] = AstJsonObject
    message: str = "Expected json object."


@dataclass
class NbtParser:
    """Parser for nbt tags."""

    array_pattern: str = r"\[[BIL];"
    curly_pattern: str = r"\{|\}"
    bracket_pattern: str = r"\[|\]"
    quoted_string_pattern: str = NBT_QUOTED_STRING_PATTERN
    number_pattern: str = r"[+-]?(?:[0-9]*?\.[0-9]+|[0-9]+\.[0-9]*?|[1-9][0-9]*|0)(?:[eE][+-]?[0-9]+)?[bslfdBSLFD]?\b"
    string_pattern: str = r"[a-zA-Z0-9._+-]+"
    colon_pattern: str = r":"
    comma_pattern: str = r","

    compound_entry_parser: Parser = field(init=False)
    list_or_array_element_parser: Parser = field(init=False)
    recursive_parser: Parser = field(init=False)

    number_suffixes: Dict[str, Type[Any]] = field(
        default_factory=lambda: {
            "b": Byte,
            "s": Short,
            "l": Long,
            "f": Float,
            "d": Double,
        }
    )

    literal_aliases: Dict[str, AstNbtValue] = field(
        default_factory=lambda: {
            "true": AstNbtBool(value=Byte(1)),
            "false": AstNbtBool(value=Byte(0)),
        }
    )

    quote_helper: QuoteHelper = field(default_factory=NbtQuoteHelper)

    def __post_init__(self):
        self.compound_entry_parser = self.parse_compound_entry
        self.list_or_array_element_parser = self
        self.recursive_parser = self

    def __call__(self, stream: TokenStream) -> AstNbt:
        with stream.syntax(
            array=self.array_pattern,
            curly=self.curly_pattern,
            bracket=self.bracket_pattern,
            quoted_string=self.quoted_string_pattern,
            number=self.number_pattern,
            string=self.string_pattern,
            colon=self.colon_pattern,
            comma=self.comma_pattern,
        ):
            curly, bracket, array, number, string, quoted_string = stream.expect(
                ("curly", "{"),
                ("bracket", "["),
                "array",
                "number",
                "string",
                "quoted_string",
            )

            if curly:
                entries: List[AstNbtCompoundEntry] = []

                for _ in stream.peek_until(("curly", "}")):
                    entries.append(self.compound_entry_parser(stream))

                    if not stream.get("comma"):
                        stream.expect(("curly", "}"))
                        break

                node = AstNbtCompound(entries=AstChildren(entries))
                return set_location(node, curly, stream.current)

            elif bracket or array:
                elements: List[AstNbt] = []

                for _ in stream.peek_until(("bracket", "]")):
                    elements.append(self.list_or_array_element_parser(stream))

                    if not stream.get("comma"):
                        stream.expect(("bracket", "]"))
                        break

                if array:
                    if array.value[1] == "B":
                        node = AstNbtByteArray(elements=AstChildren(elements))
                        element_type = Byte
                        msg = "Expected all elements to be bytes."
                    elif array.value[1] == "I":
                        node = AstNbtIntArray(elements=AstChildren(elements))
                        element_type = Int
                        msg = "Expected all elements to be integers."
                    else:
                        node = AstNbtLongArray(elements=AstChildren(elements))
                        element_type = Long
                        msg = "Expected all elements to be long integers."

                    node = set_location(node, array, stream.current)

                    for element in node.elements:
                        if isinstance(element, AstNbtValue):
                            if type(element.value) is not element_type:
                                raise element.emit_error(InvalidSyntax(msg))
                else:
                    node = AstNbtList(elements=AstChildren(elements))
                    node = set_location(node, bracket, stream.current)

                return node

            if number:
                suffix = number.value[-1].lower()

                try:
                    if suffix in self.number_suffixes:
                        value = self.number_suffixes[suffix](number.value[:-1])
                    else:
                        value = (
                            Double(number.value)
                            if "." in number.value
                            else Int(number.value)
                        )
                except (OutOfRange, ValueError):
                    value = String(number.value)

            elif string:
                alias = string.value.lower()

                if node := self.literal_aliases.get(alias):
                    return set_location(node, stream.current)
                else:
                    value = String(string.value)

            elif quoted_string:
                value = String(self.quote_helper.unquote_string(quoted_string))

            node = AstNbtValue(value=value)  # pyright: ignore[reportPossiblyUnboundVariable]
            return set_location(node, stream.current)

    def parse_compound_entry(self, stream: TokenStream) -> AstNbtCompoundEntry:
        """Parse nbt compound entry."""
        key = stream.expect_any("number", "string", "quoted_string")
        key_node = AstNbtCompoundKey(value=self.quote_helper.unquote_string(key))
        key_node = set_location(key_node, key)

        stream.expect("colon")

        value_node = self.recursive_parser(stream)

        entry_node = AstNbtCompoundEntry(key=key_node, value=value_node)
        return set_location(entry_node, key_node, value_node)


@dataclass
class NbtCompoundParser(TypeConstraint):
    """Parser for nbt compounds."""

    parser: Parser = field(default_factory=NbtParser)
    type: Union[Type[Any], Tuple[Type[Any], ...]] = AstNbtCompound
    message: str = "Expected nbt compound."


@dataclass
class AdjacentConstraint:
    """Constraint that ensures that there are no whitespace separators."""

    parser: Parser
    hint: str

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(hint=self.hint), stream.intercept("whitespace"):
            token = stream.peek()
            if not token or not token.match("hint"):
                return None
        return self.parser(stream)


@dataclass
class ResourceLocationParser:
    """Parser for resource locations."""

    def __call__(self, stream: TokenStream) -> AstResourceLocation:
        with stream.syntax(resource_location=r"#?(?:[0-9a-z_\-\.]+:)?[0-9a-z_./-]+"):
            token = stream.expect("resource_location")
            value = token.value

            if is_tag := value.startswith("#"):
                value = value[1:]

            namespace, _, path = value.rpartition(":")

            if not namespace:
                namespace = None

            node = AstResourceLocation(is_tag=is_tag, namespace=namespace, path=path)
            return set_location(node, token)


@dataclass
class NoTagConstraint:
    """Constraint that disallows resource locations refering to tags."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstResourceLocation):
            if node.is_tag:
                exc = InvalidSyntax(
                    f"Specifying a tag is not allowed {node.get_value()!r}."
                )
                raise node.emit_error(exc)

        return node


@dataclass
class CommentDisambiguation:
    """Reset the stream to the last non-comment token."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.intercept("comment"):
            consume_line_continuation(stream)

        with stream.checkpoint() as commit:
            last_comment = None

            while stream.index >= 0 and stream.current.match(
                "whitespace",
                "newline",
                "comment",
                "indent",
                "dedent",
            ):
                if stream.current.match("comment"):
                    last_comment = stream.current
                stream.index -= 1

            if last_comment:
                commit()

        stream.generator = stream.generate_tokens()

        return self.parser(stream)


@dataclass
class BlockParser:
    """Parser for minecraft blocks."""

    resource_location_parser: Parser
    block_states_parser: Parser
    data_tags_parser: Parser

    def __call__(self, stream: TokenStream) -> AstBlock:
        identifier = self.resource_location_parser(stream)

        location = identifier.location
        end_location = identifier.end_location

        if block_states := self.block_states_parser(stream):
            end_location = stream.current.end_location
        else:
            block_states = AstChildren[AstBlockState]()

        data_tags = self.data_tags_parser(stream)

        node = AstBlock(
            identifier=identifier,
            block_states=block_states,
            data_tags=data_tags,
        )
        return set_location(node, location, data_tags if data_tags else end_location)


@dataclass
class BracketedListParser:
    """Parser for bracketed list."""

    element_parser: Parser

    def __call__(self, stream: TokenStream) -> AstChildren[AstNode]:
        elements: List[AstNode] = []

        with stream.syntax(
            bracket=r"\[|\]",
            comma=r",",
        ):
            stream.expect(("bracket", "["))

            for _ in stream.peek_until(("bracket", "]")):
                elements.append(self.element_parser(stream))

                if not stream.get("comma"):
                    stream.expect(("bracket", "]"))
                    break

        return AstChildren(elements)


@dataclass
class KeyValuePairParser:
    """Parser for bracketed key-value pairs."""

    node_type: Type[AstNode]
    key_parser: Parser
    value_parser: Parser

    def __call__(self, stream: TokenStream) -> AstNode:
        with stream.syntax(equal=r"="):
            key_node = self.key_parser(stream)
            stream.expect("equal")
            value_node = self.value_parser(stream)

            entry_node = self.node_type(key=key_node, value=value_node)  # pyright: ignore[reportCallIssue]
            return set_location(entry_node, key_node, value_node)


@dataclass
class ItemParser:
    """Parser for minecraft items."""

    node_type: Type[Any]
    identifier_parser: Parser
    arguments_parser: Parser
    data_tags_parser: Parser

    def __call__(self, stream: TokenStream) -> AstNode:
        identifier = self.identifier_parser(stream)
        location = identifier.location
        end_location = identifier.end_location

        if arguments := self.arguments_parser(stream):
            end_location = stream.current.end_location
        else:
            arguments = AstChildren[AstNode]()

        data_tags = self.data_tags_parser(stream)

        node = self.node_type(
            identifier=identifier,
            arguments=arguments,
            data_tags=data_tags,
        )
        return set_location(node, location, data_tags if data_tags else end_location)


@dataclass
class RemovedDefaultItemComponentParser:
    """Parser for removed default item components."""

    key_parser: Parser

    def __call__(self, stream: TokenStream) -> AstItemRemovedDefaultComponent:
        with stream.syntax(exclamation=r"!"):
            token = stream.expect("exclamation")
            key_node = self.key_parser(stream)
        node = AstItemRemovedDefaultComponent(key=key_node)
        return set_location(node, token, key_node)


@dataclass
class ItemPredicateAlternativesParser:
    """Parser for item predicate alternatives."""

    key_parser: Parser
    value_parser: Parser

    def __call__(self, stream: TokenStream) -> AstItemPredicateAlternatives:
        alternatives: List[
            Union[AstItemPredicateTestComponent, AstItemPredicateTestPredicate]
        ] = []

        with stream.syntax(
            pipe=r"\|",
            equal=r"=",
            tilde=r"~",
            exclamation=r"!",
        ):
            while True:
                exclamation = stream.get("exclamation")
                key_node = self.key_parser(stream)

                if stream.get("tilde"):
                    test_node = AstItemPredicateTestPredicate(
                        inverted=exclamation is not None,
                        key=key_node,
                        value=self.value_parser(stream),
                    )
                else:
                    test_node = AstItemPredicateTestComponent(
                        inverted=exclamation is not None,
                        key=key_node,
                        value=(
                            self.value_parser(stream) if stream.get("equal") else None
                        ),
                    )

                alternatives.append(
                    set_location(test_node, exclamation or key_node, test_node.value)
                )

                if not stream.get("pipe"):
                    break

        node = AstItemPredicateAlternatives(alternatives=AstChildren(alternatives))
        return set_location(node, alternatives[0], alternatives[-1])


@dataclass
class NoDataTagsConstraint:
    """Constraint that disallows data tags."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstBlock):
            if node.data_tags:
                exc = InvalidSyntax("Specifying data tags not allowed.")
                raise node.emit_error(exc)
        return node


@dataclass
class BasicLiteralParser:
    """Parser for simple literal values."""

    type: Type[AstLiteral]

    def __call__(self, stream: TokenStream) -> Any:
        if not self.type.parser:
            raise ValueError(
                f"Literal node type must have an associated parser {self.type!r}."
            )
        with stream.syntax(**{self.type.parser: self.type.regex.pattern}):
            token = stream.expect(self.type.parser)
        node = self.type(value=token.value)
        return set_location(node, token)


@dataclass
class RangeParser:
    """Parser for ranges."""

    pattern: str = rf"\.\.{NUMBER_PATTERN}|{NUMBER_PATTERN}\.\.(?:{NUMBER_PATTERN})?|{NUMBER_PATTERN}"

    def __call__(self, stream: TokenStream) -> AstRange:
        with stream.syntax(range=self.pattern):
            token = stream.expect("range")
            return set_location(AstRange.from_value(token.value), token)


@dataclass
class IntegerRangeConstraint:
    """Constraint that disallows decimal ranges."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstRange):
            if isinstance(node.min, float) or isinstance(node.max, float):
                raise node.emit_error(InvalidSyntax("Expected integer range."))
        return node


@dataclass
class LengthConstraint:
    """Constraint that only allows up to a limited number of characters."""

    parser: Parser
    limit: int

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstLiteral):
            if len(node.value) > self.limit:
                exc = InvalidSyntax(f"Expected up to {self.limit} characters.")
                raise node.emit_error(exc)
        return node


@dataclass
class TimeParser(NumericParser):
    """Parser for time."""

    name: str = "time"
    pattern: str = NUMBER_PATTERN + r"[dst]?"

    def create_node(self, stream: TokenStream) -> Any:
        return set_location(AstTime.from_value(stream.current.value), stream.current)


def parse_uuid(stream: TokenStream) -> AstUUID:
    """Parse uuid."""
    with stream.syntax(uuid="-".join([r"[a-fA-F0-9]+"] * 5)):
        token = stream.expect("uuid")
    a, b, c, d, e = token.value.split("-")
    node = AstUUID(value=UUID(f"{a:>08}-{b:>04}-{c:>04}-{d:>04}-{e:>012}"))
    return set_location(node, token)


@dataclass
class SelectorParser:
    """Parser for selectors."""

    def __call__(self, stream: TokenStream) -> AstSelector:
        with stream.syntax(
            selector=r"@[praesn]\[?",
            bracket=r"\[|\]",
            comma=r",",
        ):
            token = stream.expect("selector")
            variable = token.value[1]
            location = token.location
            end_location = token.end_location

            arguments: List[AstSelectorArgument] = []

            if token.value.endswith("["):
                for _ in stream.peek_until(("bracket", "]")):
                    arguments.append(delegate("selector:argument", stream))
                    if not stream.get("comma"):
                        stream.expect(("bracket", "]"))
                        break

                end_location = stream.current.end_location

        node = AstSelector(
            variable=variable,  # pyright: ignore[reportArgumentType]
            arguments=AstChildren(arguments),
        )
        return set_location(node, location, end_location)


@dataclass
class SelectorArgumentParser:
    """Parser for selector arguments."""

    key_parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.syntax(
            equal=r"=",
            exclamation=r"!",
        ):
            key_node = self.key_parser(stream)
            argument = key_node.value

            stream.expect("equal")

            inverted = stream.get("exclamation") is not None

            hint = stream.peek()

            if hint and hint.match("comma", ("bracket", "]")):
                value_node = None
            else:
                try:
                    value_node = delegate(f"selector:argument:{argument}", stream)
                except UnrecognizedParser as exc:
                    if not exc.parser.startswith("selector:argument:"):
                        raise
                    syntax_exc = InvalidSyntax(
                        f"Invalid selector argument {argument!r}."
                    )
                    raise key_node.emit_error(syntax_exc) from exc

        node = AstSelectorArgument(inverted=inverted, key=key_node, value=value_node)
        return set_location(node, key_node, value_node)


@dataclass
class SelectorArgumentInvertConstraint:
    """Constraint that only allows inverting a specific set of arguments."""

    parser: Parser
    allow_invert: List[str]

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstSelectorArgument):
            if node.inverted and node.key.value not in self.allow_invert:
                exc = InvalidSyntax(f"Can not invert argument {node.key.value!r}.")
                raise node.emit_error(exc)
        return node


@dataclass
class SelectorArgumentNoValueConstraint:
    """Constraint that only allows a specific set of arguments to have no associated value."""

    parser: Parser
    allow_no_value: List[str]

    def __call__(self, stream: TokenStream) -> Any:
        if isinstance(node := self.parser(stream), AstSelectorArgument):
            if not node.value and node.key.value not in self.allow_no_value:
                exc = InvalidSyntax(f"Argument {node.key.value!r} must have a value.")
                raise node.emit_error(exc)
        return node


def parse_selector_scores(stream: TokenStream) -> AstSelectorScores:
    """Parse selector scores."""
    with stream.syntax(
        curly=r"\{|\}",
        objective=AstObjective.regex.pattern,
        equal=r"=",
        comma=r",",
    ):
        curly = stream.expect(("curly", "{"))

        scores: List[AstSelectorScoreMatch] = []

        for _ in stream.peek_until(("curly", "}")):
            key_node = delegate("objective", stream)
            stream.expect("equal")

            value_node = delegate("integer_range", stream)

            match_node = AstSelectorScoreMatch(key=key_node, value=value_node)
            scores.append(set_location(match_node, key_node, value_node))

            if not stream.get("comma"):
                stream.expect(("curly", "}"))
                break

        node = AstSelectorScores(scores=AstChildren(scores))
        return set_location(node, curly, stream.current)


def parse_selector_advancements(stream: TokenStream) -> AstSelectorAdvancements:
    """Parse selector advancements."""
    with stream.syntax(curly=r"\{|\}", equal=r"=", comma=r","):
        curly = stream.expect(("curly", "{"))

        advancements: List[AstSelectorAdvancementMatch] = []

        for _ in stream.peek_until(("curly", "}")):
            key_node = delegate("resource_location", stream)
            stream.expect("equal")

            if stream.get("curly"):
                predicates: List[AstSelectorAdvancementPredicateMatch] = []

                for _ in stream.peek_until(("curly", "}")):
                    pkey = delegate("advancement_predicate", stream)
                    stream.expect("equal")
                    pvalue = delegate("bool", stream)

                    predicate_node = AstSelectorAdvancementPredicateMatch(
                        key=pkey,
                        value=pvalue,
                    )
                    predicates.append(set_location(predicate_node, pkey, pvalue))

                    if not stream.get("comma"):
                        stream.expect(("curly", "}"))
                        break

                value_node = AstChildren(predicates)
            else:
                value_node = delegate("bool", stream)

            match_node = AstSelectorAdvancementMatch(key=key_node, value=value_node)
            advancements.append(set_location(match_node, key_node, value_node))

            if not stream.get("comma"):
                stream.expect(("curly", "}"))
                break

        node = AstSelectorAdvancements(advancements=AstChildren(advancements))
        return set_location(node, curly, stream.current)


@dataclass
class SelectorPlayerConstraint:
    """Constraint that disallows non player-type entities."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        if not isinstance(node, AstSelector):
            return node

        is_player = node.variable in "pras"

        for argument in node.arguments:
            if argument.key.value in ["gamemode", "level"]:
                is_player = True
            if argument.key.value == "type":
                is_player = (
                    not argument.inverted
                    and isinstance(entity_type := argument.value, AstResourceLocation)
                    and entity_type.get_canonical_value() == "minecraft:player"
                )

        if not is_player:
            exc = InvalidSyntax("Expected player-type entity selector.")
            raise node.emit_error(exc)

        return node


@dataclass
class SelectorTypeConstraint:
    """Constraint that only allows selectors that match the type property."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        properties = get_stream_properties(stream)
        selector_type = properties.get("type", "entities")

        if selector_type not in ["players", "entities"]:
            raise ValueError(f"Invalid selector type {selector_type}.")

        if selector_type == "entities":
            return self.parser(stream)

        return SelectorPlayerConstraint(self.parser)(stream)


@dataclass
class SelectorSingleConstraint:
    """Constraint that disallows selectors that target more than one entity."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        node = self.parser(stream)

        if not isinstance(node, AstSelector):
            return node

        is_single = node.variable in "prsn"

        for arg in node.arguments:
            if arg.key.value == "limit":
                is_single = arg.value == AstNumber(value=1)

        if not is_single:
            exc = InvalidSyntax("Expected entity selector targeting a single entity.")
            raise node.emit_error(exc)

        return node


@dataclass
class SelectorAmountConstraint:
    """Constraint that only allows selectors that match the amount property."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        properties = get_stream_properties(stream)

        amount = properties.get("amount", "multiple")

        if amount not in ["single", "multiple"]:
            raise ValueError(f"Invalid selector amount {amount}.")

        if amount == "multiple":
            return self.parser(stream)

        return SelectorSingleConstraint(self.parser)(stream)


@dataclass
class EntityParser:
    """Parser for entities."""

    selector_parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        """Parse entity."""
        with stream.syntax(literal=AstLiteral.regex.pattern):
            hint = stream.peek()

        if hint:
            if hint.value.startswith("@"):
                return self.selector_parser(stream)

            if hint.value.count("-") == 4:
                with stream.alternative():
                    return delegate("uuid", stream)

        return delegate("player_name", stream)


def parse_wildcard(stream: TokenStream) -> AstWildcard:
    """Parse wildcard."""
    with stream.syntax(wildcard=r"\*"):
        token = stream.expect("wildcard")
        return set_location(AstWildcard(), token)


def parse_message(stream: TokenStream) -> AstMessage:
    """Parse message."""
    multiline = get_stream_multiline(stream)

    with stream.intercept("whitespace"):
        stream.get("whitespace")

    with (
        stream.intercept("newline"),
        stream.syntax(
            selector=r"@[praesn]",
            text=r"[^\n@]+",
        ),
    ):
        fragments: List[Any] = []

        while True:
            selector, text = stream.expect("selector", "text")

            if selector:
                stream.index -= 1
                with stream.ignore("newline") if multiline else nullcontext():
                    with stream.syntax(text=None):
                        fragments.append(delegate("selector", stream))
            elif text:
                text_node = AstMessageText(value=text.value)
                text_node = set_location(text_node, text)
                fragments.append(text_node)

            with stream.syntax(text=None):
                if consume_line_continuation(stream):
                    whitespace_node = AstMessageText(value=" ")
                    whitespace_node = set_location(
                        whitespace_node,
                        stream.current.end_location.with_horizontal_offset(-1),
                        stream.current.end_location,
                    )
                    fragments.append(whitespace_node)
                else:
                    with stream.checkpoint():
                        if stream.get("newline", "eof"):
                            break

    node = AstMessage(fragments=AstChildren(fragments))
    return set_location(node, fragments[0], fragments[-1])


@dataclass
class NbtPathParser:
    """Parser for nbt paths."""

    integer_parser: Parser = field(
        default_factory=lambda: IntegerConstraint(NumericParser())
    )
    nbt_compound_parser: Parser = field(default_factory=NbtCompoundParser)

    quote_helper: QuoteHelper = field(
        default_factory=lambda: QuoteHelper(
            escape_sequences={
                r"\\": "\\",
            }
        )
    )

    def __call__(self, stream: TokenStream) -> AstNbtPath:
        components: List[Any] = []

        with stream.syntax(
            dot=r"\.",
            curly=r"\{|\}",
            bracket=r"\[|\]",
            quoted_string=r'"(?:\\.|[^\\\n])*?"' "|" r"'(?:\\.|[^\\\n])*?'",
            string=r"(?:[0-9a-z_\-]+:)?[a-zA-Z0-9_+-]+",
        ):
            components.extend(self.parse_modifiers(stream))

            while not components or stream.get("dot"):
                quoted_string, string = stream.expect("quoted_string", "string")

                if quoted_string:
                    component_node = AstNbtPathKey(
                        value=self.quote_helper.unquote_string(quoted_string),
                    )
                    components.append(set_location(component_node, quoted_string))
                elif string:
                    component_node = AstNbtPathKey(value=string.value)
                    components.append(set_location(component_node, string))

                components.extend(self.parse_modifiers(stream))

        if not components:
            raise stream.emit_error(InvalidSyntax("Empty nbt path not allowed."))

        node = AstNbtPath(components=AstChildren(components))
        return set_location(node, components[0], components[-1])

    def parse_modifiers(self, stream: TokenStream) -> Iterator[Any]:
        """Parse named tag modifiers."""
        hint = stream.peek()

        if hint and hint.match(("curly", "{")):
            yield self.nbt_compound_parser(stream)
            return

        while bracket := stream.get(("bracket", "[")):
            index = None

            hint = stream.peek()
            if hint and hint.match(("curly", "{")):
                index = self.nbt_compound_parser(stream)
            elif hint and not hint.match(("bracket", "]")):
                index = self.integer_parser(stream)

            close_bracket = stream.expect(("bracket", "]"))

            subscript_node = AstNbtPathSubscript(index=index)
            yield set_location(subscript_node, bracket, close_bracket)


@dataclass
class ParticleParser:
    resource_location_parser: Parser
    generic_parameters_parser: Parser

    def __call__(self, stream: TokenStream) -> AstParticle:
        """Parse particle."""
        parameters = None

        if isinstance(
            name := self.resource_location_parser(stream), AstResourceLocation
        ):
            with stream.syntax(hint=r"\{"), stream.intercept("whitespace"):
                token = stream.peek()

            if token and token.match("hint"):
                parameters = self.generic_parameters_parser(stream)
            else:
                try:
                    parameters = delegate(
                        f"particle:{name.get_canonical_value()}", stream
                    )
                except UnrecognizedParser as exc:
                    if not exc.parser.startswith("particle:"):
                        raise

        node = AstParticle(name=name, parameters=parameters)
        return set_location(node, name, parameters if parameters else name)


def parse_macro_line(stream: TokenStream) -> AstMacroLine:
    """Parse macro line."""
    with stream.syntax(macro_line=r"\$"):
        token = stream.expect("macro_line")

    has_variable = False
    arguments: List[AstNode] = []

    with (
        stream.intercept("newline", "whitespace"),
        stream.syntax(open_variable=r"\$\("),
    ):
        for _ in stream.peek_until("newline", "eof"):
            if open_variable := stream.get("open_variable"):
                node = delegate("macro_line_variable", stream)
                with stream.syntax(close_variable=r"\)"):
                    close_variable = stream.expect("close_variable")
                arguments.append(set_location(node, open_variable, close_variable))
                has_variable = True
            else:
                arguments.append(delegate("macro_line_text", stream))

    if not has_variable:
        exc = InvalidSyntax("Expected at least one macro line variable.")
        raise set_location(exc, token, stream.current)

    node = AstMacroLine(arguments=AstChildren(arguments))
    return set_location(node, token, arguments[-1])


@dataclass
class AggregateParser:
    """Parser that creates a node by parsing the given fields one after the other."""

    type: Type[AstNode]
    fields: Dict[str, Parser]

    def __call__(self, stream: TokenStream) -> Any:
        values = [(name, parser(stream)) for name, parser in self.fields.items()]
        node = self.type(**dict(values))
        return set_location(node, values[0][1], values[-1][1])


@dataclass
class SingleLineConstraint:
    """Constraint that prevents parsing across multiple lines."""

    parser: Parser

    def __call__(self, stream: TokenStream) -> Any:
        with stream.intercept("newline"):
            return self.parser(stream)


@dataclass
class AlternativeParser:
    """Parser that tries the given alternatives in order."""

    parsers: Sequence[Parser]

    def __call__(self, stream: TokenStream) -> Any:
        for parser, alternative in stream.choose(*self.parsers):
            with alternative:
                return parser(stream)



__version__ = "0.102.0-beta.3"

from .api import *  # noqa: F403
from .ast import *  # noqa: F403
from .config import *  # noqa: F403
from .database import *  # noqa: F403
from .diagnostic import *  # noqa: F403
from .dispatch import *  # noqa: F403
from .error import *  # noqa: F403
from .parse import *  # noqa: F403
from .plugin import *  # noqa: F403
from .preprocess import *  # noqa: F403
from .prototype import *  # noqa: F403
from .resources import *  # noqa: F403
from .serialize import *  # noqa: F403
from .spec import *  # noqa: F403



"""Plugin for baking static macro invocations."""

__all__ = [
    "bake_macros",
    "BakeMacrosOptions",
    "MacroBaker",
    "MacroInvocation",
    "MacroInvocationArgument",
    "MacroBakerParseCallback",
]


from dataclasses import dataclass, field, replace
from typing import Any, Dict, List, Protocol, Set
from weakref import WeakKeyDictionary

from beet import (
    Context,
    Function,
    Generator,
    ListOption,
    PluginOptions,
    TextFileBase,
    configurable,
)
from beet.core.utils import required_field
from nbtlib import Numeric, String
from pathspec import PathSpec
from tokenstream import INITIAL_LOCATION, Preprocessor, SourceLocation, set_location

from mecha import (
    AstChildren,
    AstCommand,
    AstMacroLine,
    AstMacroLineText,
    AstMacroLineVariable,
    AstNbtCompound,
    AstNbtCompoundEntry,
    AstNbtValue,
    AstResourceLocation,
    AstRoot,
    CompilationDatabase,
    Diagnostic,
    DiagnosticCollection,
    DiagnosticError,
    Mecha,
    Serializer,
    Visitor,
    rule,
)


class BakeMacrosOptions(PluginOptions):
    match: ListOption[str] = ListOption.model_validate(["*"])
    macro: ListOption[str] = ListOption.model_validate(["*"])


def beet_default(ctx: Context):
    ctx.require(bake_macros)


@configurable(validator=BakeMacrosOptions)
def bake_macros(ctx: Context, opts: BakeMacrosOptions):
    """Plugin for baking static macro invocations."""
    mc = ctx.inject(Mecha)
    mc.steps.insert(
        mc.steps.index(mc.transform) + 1,
        MacroBaker(
            match_spec=PathSpec.from_lines("gitwildmatch", opts.match.entries()),
            macro_spec=PathSpec.from_lines("gitwildmatch", opts.macro.entries()),
            database=mc.database,
            serialize=mc.serialize,
            generate=ctx.generate,
            parse_callback=mc.parse,
        ),
    )


class MacroBakerParseCallback(Protocol):
    """Callback required by the macro baker for parsing."""

    def __call__(
        self,
        source: TextFileBase[Any],
        *,
        preprocessor: Preprocessor,
    ) -> AstRoot: ...


@dataclass
class MacroInvocationArgument:
    """Argument provided to a static macro invocation."""

    entry: AstNbtCompoundEntry
    stringified: str


@dataclass
class MacroInvocation:
    """Static macro invocation."""

    file_instance: TextFileBase[Any]
    macro_path: str
    node: AstCommand
    arguments: Dict[str, MacroInvocationArgument]

    diagnostics: DiagnosticCollection = field(default_factory=DiagnosticCollection)
    arguments_usage: Dict[str, List[AstMacroLine]] = field(default_factory=dict)
    arguments_errors: Dict[str, Set[str]] = field(default_factory=dict)

    def add_usage(self, argument: str, node: AstMacroLine):
        self.arguments_usage.setdefault(argument, []).append(node)

    def add_error(self, argument: str, error_code: str):
        self.arguments_errors.setdefault(argument, set()).add(error_code)


@dataclass(eq=False)
class MacroBaker(Visitor):
    """Visitor for baking static macro invocations."""

    match_spec: PathSpec = required_field()
    macro_spec: PathSpec = required_field()

    database: CompilationDatabase = required_field()
    serialize: Serializer = required_field()
    generate: Generator = required_field()
    parse_callback: MacroBakerParseCallback = required_field()

    invocations: WeakKeyDictionary[TextFileBase[Any], MacroInvocation] = field(
        default_factory=WeakKeyDictionary
    )

    @rule(AstRoot)
    def bake_macros(self, node: AstRoot):
        invocation = self.invocations.get(self.database.current)

        result: List[AstCommand] = []
        modified = False

        for command in node.commands:
            if isinstance(command, AstMacroLine) and invocation:
                baked_macro_line = self.bake_macro_line(command, invocation)
                modified |= baked_macro_line is not command
                command = baked_macro_line

            args = command.arguments
            stack: List[AstCommand] = [command]

            while args and isinstance(subcommand := args[-1], AstCommand):
                stack.append(subcommand)
                args = subcommand.arguments

            last = stack[-1]

            if last.identifier == "function:name:arguments":
                last = self.bake_macro_invocation(last)
                if last is not stack.pop():
                    for prefix in reversed(stack):
                        args = AstChildren([*prefix.arguments[:-1], last])
                        last = replace(prefix, arguments=args)
                    modified = True
                    command = last

            result.append(command)

        if invocation:
            invocation_compilation_unit = self.database[invocation.file_instance]

            for name, errors in invocation.arguments_errors.items():
                diagnostics = DiagnosticCollection(
                    hint=invocation_compilation_unit.resource_location,
                    filename=invocation_compilation_unit.filename,
                    file=invocation.file_instance,
                )

                if "missing" in errors:
                    d = Diagnostic(
                        "error",
                        f'Missing argument "{name}" when baking macro "{invocation.macro_path}".',
                    )
                    diagnostics.add(set_location(d, invocation.node.arguments[1]))
                if "syntax" in errors:
                    d = Diagnostic(
                        "warn",
                        f'Argument "{name}" triggers a syntax error when baking macro "{invocation.macro_path}".',
                    )
                    diagnostics.add(set_location(d, invocation.arguments[name].entry))
                if "zero" in errors:
                    d = Diagnostic(
                        "error",
                        f'Argument "{name}" results in macro line with no command when baking macro "{invocation.macro_path}".',
                    )
                    diagnostics.add(set_location(d, invocation.arguments[name].entry))
                if "many" in errors:
                    d = Diagnostic(
                        "error",
                        f'Argument "{name}" results in macro line with more than one command when baking macro "{invocation.macro_path}".',
                    )
                    diagnostics.add(set_location(d, invocation.arguments[name].entry))

                macro_lines = "".join(
                    f"\n> {self.serialize(macro_line)}"
                    for macro_line in invocation.arguments_usage[name]
                )

                for exc in diagnostics.exceptions:
                    exc.notes.append(f"Required in macro line{macro_lines}")

                yield diagnostics

            for name, invocation_argument in invocation.arguments.items():
                if name not in invocation.arguments_usage:
                    d = Diagnostic(
                        "warn",
                        f'Unused argument "{name}" when baking macro "{invocation.macro_path}".',
                        hint=invocation_compilation_unit.resource_location,
                        filename=invocation_compilation_unit.filename,
                        file=invocation.file_instance,
                    )
                    yield set_location(d, invocation_argument.entry)

            yield invocation.diagnostics

        if modified:
            node = replace(node, commands=AstChildren(result))

        return node

    def bake_macro_line(self, node: AstMacroLine, invocation: MacroInvocation):
        result: List[str] = []
        source_mappings: list[SourceLocation] = []
        preprocessed_mappings: list[SourceLocation] = []
        preprocessed_location = INITIAL_LOCATION

        used_variables: List[AstMacroLineVariable] = []
        valid = True

        for argument in node.arguments:
            if isinstance(argument, AstMacroLineVariable):
                invocation.add_usage(argument.value, node)

                invocation_argument = invocation.arguments.get(argument.value)
                if not invocation_argument:
                    invocation.add_error(argument.value, "missing")
                    valid = False
                    continue

                used_variables.append(argument)

                result.append(invocation_argument.stringified)
                source_mappings.append(argument.location)
                preprocessed_mappings.append(preprocessed_location)
                preprocessed_location = preprocessed_location.skip_over(
                    invocation_argument.stringified
                )

            elif isinstance(argument, AstMacroLineText):
                result.append(argument.value)
                source_mappings.append(argument.location)
                preprocessed_mappings.append(preprocessed_location)
                preprocessed_location = preprocessed_location.skip_over(argument.value)

        if not valid:
            return node

        source = "".join(result)

        try:
            root = self.parse_callback(
                self.database.current,
                preprocessor=lambda _: (source, source_mappings, preprocessed_mappings),
            )

        except DiagnosticError as exc:
            syntax_error = exc.diagnostics.exceptions[0]
            syntax_error.notes.append(f"Failed to parse baked macro line\n> {source}")
            invocation.diagnostics.add(syntax_error)

            for argument in used_variables:
                if (
                    syntax_error.location < argument.end_location
                    and argument.location < syntax_error.end_location
                ):
                    invocation.add_error(argument.value, "syntax")

            return node

        else:
            if len(root.commands) == 1:
                return root.commands[0]

            for argument in used_variables:
                invocation.add_error(
                    argument.value, "many" if root.commands else "zero"
                )

            return node

    def bake_macro_invocation(self, node: AstCommand):
        current_file = self.database.current
        path = self.database[current_file].resource_location

        if (
            path
            and self.match_spec.match_file(path)
            and len(node.arguments) == 2
            and isinstance(name := node.arguments[0], AstResourceLocation)
            and self.macro_spec.match_file(macro_path := name.get_canonical_value())
            and isinstance(arguments := node.arguments[1], AstNbtCompound)
            and (macro_file := self.database.index.get(macro_path))
        ):
            invocation_arguments = {
                entry.key.value: MacroInvocationArgument(
                    entry=entry,
                    stringified=(
                        str(nbt_tag.unpack())
                        if isinstance(entry.value, AstNbtValue)
                        and isinstance(
                            nbt_tag := entry.value.evaluate(), (String, Numeric)
                        )
                        else self.serialize(entry.value)
                    ),
                )
                for entry in arguments.entries
            }

            invocation_path = self.generate.format(
                macro_path + "/baked_{hash}",
                hash=sorted(
                    (k, v.stringified) for k, v in invocation_arguments.items()
                ),
            )

            if not self.database.index.get(invocation_path):
                invocation = Function(original=macro_file.original)
                self.generate(invocation_path, invocation)
                self.database[invocation] = replace(
                    self.database[macro_file],
                    resource_location=invocation_path,
                )
                self.database.enqueue(invocation, self.database.step)
                self.invocations[invocation] = MacroInvocation(
                    file_instance=current_file,
                    macro_path=macro_path,
                    node=node,
                    arguments=invocation_arguments,
                )

            name = set_location(AstResourceLocation.from_value(invocation_path), name)
            node = replace(
                node,
                identifier="function:name",
                arguments=AstChildren([name]),
            )

        return node
